import time

from mock import patch
from transifex.native import TxNative
from transifex.native.daemon import DaemonicThread


class TestFetchingDaemon(object):

    @patch('transifex.native.daemon.tx')
    def test_daemon_starts(self, patched_tx):
        tx = TxNative()
        tx.init(['en', 'el'], 'some:token', 'https://some.host')

        # the `interval` we will be using
        interval = 1

        daemon = DaemonicThread()

        # nothing is running
        assert not daemon.is_daemon_running(log_errors=False)

        # start deamon
        daemon.start_daemon(interval=interval)

        # test that thread has started
        time.sleep(interval * 2)
        assert daemon.is_daemon_running(log_errors=False)

        # test that `fetch_translations` has been called
        assert patched_tx.fetch_translations.call_count > 0

        daemon.stop_daemon()
        assert not daemon.is_daemon_running()

    @patch('transifex.native.daemon.tx')
    @patch('transifex.native.daemon.logger')
    def test_daemon_exception(self, patched_logger, patched_tx):

        patched_tx.fetch_translations.side_effect = Exception(
            'Something went wrong')

        tx = TxNative()
        tx.init(['en', 'el'], 'some:token', 'https://some.host')

        daemon = DaemonicThread()

        interval = 1
        daemon.start_daemon(interval=1)
        time.sleep(interval * 2)
        assert daemon.is_daemon_running(log_errors=False)

        patched_logger.exception.assert_called_with(
            'Fetching daemon exception: Something went wrong'
        )
        daemon.stop_daemon()



import pytest
from mock import patch
from transifex.native.tools.migrations.mark import (
    MARK_PROOFREAD_FILE, MARK_PROOFREAD_STRING, MarkLowConfidenceFilesPolicy,
    MarkLowConfidenceStringsPolicy, MarkPolicy, NoopMarkPolicy)
from transifex.native.tools.migrations.models import (Confidence,
                                                      FileMigration,
                                                      StringMigration)


def test_base_class_policy_returns_false():
    policy = NoopMarkPolicy()
    assert policy.mark_file(_file()) is False
    assert policy.mark_string(_string()) is False


@patch('transifex.native.tools.migrations.mark.Color.echo')
@patch('transifex.native.tools.migrations.mark.mark_string')
def test_low_file_policy_marks_low_confidence_files(mock_mark_string,
                                                    mock_echo):
    # Create a file migration with 3 strings,
    # the middle one with low confidence
    policy = MarkLowConfidenceFilesPolicy()
    policy.set_comment_format('# {}')
    file_migration = _file()
    file_migration.add_string(_string(Confidence.HIGH))
    file_migration.add_string(_string(Confidence.LOW))
    file_migration.add_string(_string(Confidence.HIGH))

    # A proofread mark should have been added before the first string,
    # as the files policy adds a mark at the file top
    policy.mark_file(file_migration)
    mock_mark_string.assert_called_once_with(
        file_migration.strings[0],
        '# {}',
        MARK_PROOFREAD_FILE,
    )

    # Calls to mark_string() of a files policy should not mark anything
    mock_mark_string.reset_mock()
    policy.mark_string(file_migration.strings[1])
    assert mock_mark_string.called is False


@patch('transifex.native.tools.migrations.mark.Color.echo')
@patch('transifex.native.tools.migrations.mark.mark_string')
def test_low_string_policy_marks_low_confidence_strings(mock_mark_string,
                                                        mock_echo):
    # Create a file migration with 3 strings,
    # two of which have low confidence
    policy = MarkLowConfidenceStringsPolicy()
    policy.set_comment_format('<!-- {} -->')
    file_migration = _file()
    file_migration.add_string(_string(Confidence.HIGH))
    file_migration.add_string(_string(Confidence.LOW))
    file_migration.add_string(_string(Confidence.LOW))

    policy.mark_file(file_migration)
    assert mock_mark_string.called is False

    for string_migration in file_migration.strings:
        policy.mark_string(string_migration)

    assert mock_mark_string.call_args_list[0][0] == (
        file_migration.strings[1],
        '<!-- {} -->',
        MARK_PROOFREAD_STRING,
    )
    assert mock_mark_string.call_args_list[1][0] == (
        file_migration.strings[2],
        '<!-- {} -->',
        MARK_PROOFREAD_STRING,
    )


def test_set_comment_format_exception_for_wrong_format():
    # An exception should be raised if the given format does not include {}
    policy = MarkPolicy()
    with pytest.raises(ValueError):
        policy.set_comment_format('{')


def _string(confidence=Confidence.HIGH):
    """Return a sample StringMigration object for testing."""
    return StringMigration('original', 'new', confidence)


def _file():
    """Return a sample FileMigration object for testing."""
    return FileMigration('filename', 'content')



from __future__ import absolute_import, unicode_literals

import os

import transifex.native.tools.migrations.gettext as gettext
from django.conf import settings
from django.core.management.utils import handle_extensions
from transifex.native.django.management.common import TranslatableFile
from transifex.native.django.management.utils.base import (CommandMixin,
                                                           pretty_options)
from transifex.native.django.tools.migrations.templatetags import \
    DjangoTagMigrationBuilder
from transifex.native.tools.migrations.execution import (MARK_POLICY_OPTIONS,
                                                         REVIEW_POLICY_OPTIONS,
                                                         SAVE_POLICY_OPTIONS,
                                                         MigrationExecutor,
                                                         migrate_text)
from transifex.native.tools.migrations.gettext import (GettextMethods,
                                                       GettextMigrationBuilder)

MIGRATE_EXTENSIONS = ['html', 'txt', 'py']


# These are the functions + arguments of the gettext wrappers
# provided in Django
GETTEXT_FUNCTIONS = {
    gettext.GETTEXT: (
        'django.utils.translation.gettext',
        [('message', gettext.KEYWORD_STRING)],
    ),
    gettext.UGETTEXT: (
        'django.utils.translation.ugettext',
        [('message', gettext.KEYWORD_STRING)],
    ),
    gettext.NGETTEXT: (
        'django.utils.translation.ngettext',
        [
            ('singular', gettext.KEYWORD_ONE),
            ('plural', gettext.KEYWORD_OTHER),
            ('number', gettext.KEYWORD_CNT),
        ]
    ),
    gettext.UNGETTEXT: (
        'django.utils.translation.ungettext',
        [
            ('singular', gettext.KEYWORD_ONE),
            ('plural', gettext.KEYWORD_OTHER),
            ('number', gettext.KEYWORD_CNT),
        ]
    ),
    gettext.PGETTEXT: (
        'django.utils.translation.pgettext',
        [
            ('context', gettext.KEYWORD_CONTEXT),
            ('message', gettext.KEYWORD_STRING),
        ]
    ),
    gettext.NPGETTEXT: (
        'django.utils.translation.npgettext',
        [
            ('context', gettext.KEYWORD_CONTEXT),
            ('singular', gettext.KEYWORD_ONE),
            ('plural', gettext.KEYWORD_OTHER),
            ('number', gettext.KEYWORD_CNT),
        ]
    ),
}
# Add lazy variants with identical arguments as the corresponding non-lazy ones
for func_name, lazy_func_name in gettext.LAZY_MAPPING.items():
    item = GETTEXT_FUNCTIONS[func_name]
    GETTEXT_FUNCTIONS[lazy_func_name] = (
        item[0] + gettext.LAZY_SUFFIX, item[1]
    )

# This is the import statement that will replace gettext imports
T_IMPORT = 'from transifex.native.django import {}'


class Migrate(CommandMixin):
    """Migrate files using the Django i18n syntax to Transifex Native syntax."""

    def add_arguments(self, subparsers):
        parser = subparsers.add_parser(
            'migrate',
            help=("Migrate files using the Django i18n syntax to Transifex "
                  "Native syntax")
        )
        parser.add_argument(
            '--file', '-f', dest='files', action='append',
            help=('The relative paths of the files to migrate. Separate '
                  'multiple paths with commas, or use -f multiple times.'),
        )
        parser.add_argument(
            '--path', '-p', dest='path',
            help='The path of the files to migrate. Finds files recursively.',
        )
        parser.add_argument(
            '--text', '-t', dest='text', default='',
            help='If set, the migration command will display the result '
                 'of converting the given text to Transifex Native syntax, '
                 'and then exit.',
        )
        parser.add_argument(
            '--save', dest='save_policy', default='new',
            help=('Determines where the migrated content will be saved: \n' +
                  pretty_options(SAVE_POLICY_OPTIONS)),
        )
        parser.add_argument(
            '--review', dest='review_policy', default='file',
            help=('Determines where the migrated content will be saved: \n' +
                  pretty_options(REVIEW_POLICY_OPTIONS)),
        )
        parser.add_argument(
            '--mark', dest='mark_policy', default='none',
            help=('Determines if anything gets marked for proofreading: \n' +
                  pretty_options(MARK_POLICY_OPTIONS)),
        )
        parser.add_argument(
            '--verbose', '-v', action='store_true',
            dest='verbose_output', default=False,
            help=('Verbose output'),
        )

    def handle(self, *args, **options):
        self.domain = options['domain']
        self.verbose_output = options['verbose_output']
        self.ignore_patterns = []
        self.path = options['path']
        self.files = set(options['files'] or [])
        exts = MIGRATE_EXTENSIONS
        self.extensions = handle_extensions(exts)
        self.stats = {
            'processed_files': 0, 'migrations': [], 'saved': [], 'errors': [],
        }

        # Create a reusable migrator for templates code
        self.django_migration_builder = DjangoTagMigrationBuilder()
        self.gettext_migration_builder = GettextMigrationBuilder(
            methods=GettextMethods(**GETTEXT_FUNCTIONS),
            import_statement=T_IMPORT,
        )

        # -- Text mode: simply transform the given text and exit
        text = options['text']
        if text:
            migrate_text(text, self._migrate_text)
            return

        # -- File mode: read all files based on the given options and migrate
        # each of them
        self.executor = MigrationExecutor(
            options, file_migrator_func=self._migrate_file,
        )

        # Show an intro message
        self.executor.show_intro()

        # If specific files are defined, use those
        if self.files:
            dirpath = os.getcwd()
            files = [
                TranslatableFile(dirpath, filename)
                for filename in self.files
            ]
        # Else search the file path for supported files
        else:
            files = self._find_files(self.path, 'migrate')

        # Execute the migration
        self.executor.migrate_files(files)

    def _migrate_text(self, text):
        """Create a migration to Native syntax for the given string.

        Supports both Python files and Django template files.

        :param unicode text: the code string
        :return: an object with the migration info
        :rtype: FileMigration
        """
        # Template code
        if '{%' in text:
            return self.django_migration_builder.build_migration(text, '')
        # Python code
        else:
            return self.gettext_migration_builder.build_migration(text, '')

    def _migrate_file(self, translatable_file):
        """Create a migration to Native syntax for the given file.

        Supports both Python files and Django template files.

        :param TranslatableFile translatable_file: the file to search
        :return: an object with the migration info
        :rtype: FileMigration
        """
        self.verbose('Processing file %s in %s' % (
            translatable_file.file, translatable_file.dirpath
        ))
        encoding = (
            settings.FILE_CHARSET if (
                hasattr(settings, 'FILE_CHARSET')
                and self.settings_available
            )
            else 'utf-8'
        )
        try:
            src_data = self._read_file(translatable_file.path, encoding)
        except UnicodeDecodeError as e:
            self.verbose(
                'UnicodeDecodeError: skipped file %s in %s (reason: %s)' % (
                    translatable_file.file, translatable_file.dirpath, e,
                )
            )
            return None

        _, extension = os.path.splitext(translatable_file.file)

        # Python file
        if extension == '.py':
            return self.gettext_migration_builder.build_migration(
                src_data, translatable_file.path,
            )

        # Template file
        return self.django_migration_builder.build_migration(
            src_data, translatable_file.path, encoding,
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from transifex.native.django.tools.migrations.templatetags import \
    DjangoTagMigrationBuilder

DJANGO_TEMPLATE = """
{% extends 'base.html' %}
{% load i18n %}

{# Translators: This is an amazing comment #}
{% trans "Hello!" %}

{% comment %}Translators: This is another amazing comment{% endcomment %}
{% trans "May" context "month name" %}

{% with first=events.first %}
    {% with second=events.second %}
        {% blocktrans with third=events.third|title another='foo' %}
        The following events were found:
          1. {{ first }}
          2. {{ second }}
          3. {{ third }}
        And this is {{ another }}.
        {% endblocktrans %}
    {% endwith %}
{% endwith %}

{% blocktrans count counter='something'|length %}
There is only one {{ name }} object.
{% plural %}
There are {{ counter }} {{ name }} objects.
{% endblocktrans %}

{% blocktrans trimmed %}
  First sentence.
  Second paragraph.
{% endblocktrans %}

<a href="{{ url }}">Text</a>
{{ _(some_other_var) }}
{% trans some_other_var|some_filter %}
{% trans "try as" as try %}
{% trans "try with <xml>xml</xml>" %}
{% blocktrans %}
    try with <xml>xml</xml>
{% endblocktrans %}
{% blocktrans with organization_name=organization.name %}Help "{{organization_name}}" translate content {% endblocktrans %}
{% blocktrans %}To download, please click <a href="https://{{ site_domain }}{{ url }}">here</a>.{% endblocktrans %}

{% comment "Optional note" %}
    <p>Commented out text with {{ create_date|date:"c" }}</p>
{% endcomment %}
{# This is not related to translations #}
{% something %}
{% comment %}This is a non-translator comment{% endcomment %}
{% comment %}Translators: This is an orphan translator comment{% endcomment %}
"""

TRANSIFEX_TEMPLATE = """
{% extends 'base.html' %}
{% load i18n %}
{% load transifex %}


{% t "Hello!" _comment="This is an amazing comment" %}


{% t "May" _comment="This is another amazing comment" _context="month name" %}

{% with first=events.first %}
    {% with second=events.second %}
        {% t another='foo' third=events.third|title %}
        The following events were found:
          1. {first}
          2. {second}
          3. {third}
        And this is {another}.
        {% endt %}
    {% endwith %}
{% endwith %}

{% t counter='something'|length %}
{counter, plural, one {
There is only one {name} object.
} other {
There are {counter} {name} objects.
}}
{% endt %}

{% t |trimmed %}
  First sentence.
  Second paragraph.
{% endt %}

<a href="{{ url }}">Text</a>
{% t some_other_var %}
{% t some_other_var|some_filter %}
{% t "try as" as try %}
{% ut "try with <xml>xml</xml>" %}
{% ut %}
    try with <xml>xml</xml>
{% endut %}
{% ut 'Help "{organization_name}" translate content ' organization_name=organization.name %}
{% ut 'To download, please click <a href="https://{site_domain}{url}">here</a>.' %}

{% comment "Optional note" %}
    <p>Commented out text with {{ create_date|date:"c" }}</p>
{% endcomment %}
{# This is not related to translations #}
{% something %}
{% comment %}This is a non-translator comment{% endcomment %}
"""


def test_compiled_string_is_expected():
    """Test all known migration cases."""
    builder = DjangoTagMigrationBuilder()
    file_migration = builder.build_migration(DJANGO_TEMPLATE)
    compiled = file_migration.compile()
    assert compiled == TRANSIFEX_TEMPLATE

    # Make sure the migration is idempotent
    file_migration = builder.build_migration(compiled)
    assert file_migration.compile() == TRANSIFEX_TEMPLATE



KEY_DEVELOPER_COMMENT = '_comment'
KEY_CHARACTER_LIMIT = '_charlimit'
KEY_TAGS = '_tags'
KEY_CONTEXT = '_context'
KEY_KEY = '_key'
KEY_OCCURRENCES = '_occurrences'

ALL_KEYS = [
    KEY_CONTEXT,
    KEY_DEVELOPER_COMMENT,
    KEY_CHARACTER_LIMIT,
    KEY_TAGS,
    KEY_OCCURRENCES,
    KEY_KEY,
]



# -*- coding: utf-8 -*-

import pytest
from mock import MagicMock, patch
from transifex.common.utils import generate_key
from transifex.native.cache import MemoryCache
from transifex.native.cds import TRANSIFEX_CDS_HOST
from transifex.native.core import NotInitializedError, TxNative
from transifex.native.parsing import SourceString
from transifex.native.rendering import (PseudoTranslationPolicy,
                                        SourceStringPolicy)
from transifex.native.settings import parse_error_policy


class TestSourceString(object):
    """Tests the functionality of the SourceString class."""

    def test_default_values(self):
        string = SourceString('something')
        assert string.string == 'something'
        assert string.context is None
        assert string.meta == {}
        assert string.developer_comment is None
        assert string.character_limit is None
        assert string.tags == []

    def test_custom_meta(self):
        string = SourceString(
            'something',
            _context='one,two,three',
            _comment='A crucial comment',
            _charlimit=33,
            _tags=' t1,t2 ,  t3',
            custom='custom',
        )
        assert string.string == 'something'
        assert string.context == ['one', 'two', 'three']
        assert string.developer_comment == 'A crucial comment'
        assert string.character_limit == 33
        assert string.tags == ['t1', 't2', 't3']
        assert string.meta.get('custom') is None

    def test_tag_list(self):
        string = SourceString(
            'something',
            _tags=['t1', 't2', 't3'],
        )
        assert string.tags == ['t1', 't2', 't3']


class TestNative(object):
    """Tests the TxNative class."""

    def _get_tx(self, **kwargs):
        mytx = TxNative()
        mytx.init(['en', 'el'], 'cds_token', **kwargs)
        return mytx

    def test_uninitialized(self):
        mytx = TxNative()
        with pytest.raises(NotInitializedError):
            mytx.translate('string', 'en')
        with pytest.raises(NotInitializedError):
            mytx.fetch_translations()
        with pytest.raises(NotInitializedError):
            mytx.push_source_strings([], False)

    def test_default_init(self):
        mytx = self._get_tx()
        assert mytx.initialized is True
        assert mytx._languages == ['en', 'el']
        assert isinstance(mytx._missing_policy, SourceStringPolicy)
        assert isinstance(mytx._cache, MemoryCache)
        assert mytx._cds_handler.token == 'cds_token'
        assert mytx._cds_handler.host == TRANSIFEX_CDS_HOST

    def test_custom_init(self):
        missing_policy = PseudoTranslationPolicy()
        mytx = self._get_tx(cds_host='myhost', missing_policy=missing_policy)
        assert mytx.initialized is True
        assert mytx._languages == ['en', 'el']
        assert mytx._missing_policy == missing_policy
        assert isinstance(mytx._cache, MemoryCache)
        assert mytx._cds_handler.token == 'cds_token'
        assert mytx._cds_handler.host == 'myhost'

    @patch('transifex.native.core.StringRenderer.render')
    def test_translate_source_language_reaches_renderer(self, mock_render):
        mytx = self._get_tx()
        mytx.translate('My String', 'en', is_source=True)
        mock_render.assert_called_once_with(
            source_string='My String',
            string_to_render='My String',
            language_code='en',
            escape=True,
            missing_policy=mytx._missing_policy,
            params={},
        )

    @patch('transifex.native.core.MemoryCache.get')
    @patch('transifex.native.core.StringRenderer.render')
    def test_translate_target_language_missing_reaches_renderer(self, mock_render,
                                                                mock_cache):
        mock_cache.return_value = None
        mytx = self._get_tx()
        mytx.translate('My String', 'en', is_source=False)
        mock_cache.assert_called_once_with(
            generate_key(string='My String'), 'en')
        mock_render.assert_called_once_with(
            source_string='My String',
            string_to_render=None,
            language_code='en',
            escape=True,
            missing_policy=mytx._missing_policy,
            params={},
        )

    def test_translate_target_language_missing_reaches_missing_policy(self):
        missing_policy = MagicMock()
        mytx = self._get_tx(missing_policy=missing_policy)
        mytx.translate('My String', 'en', is_source=False)
        missing_policy.get.assert_called_once_with('My String')

    @patch('transifex.native.core.StringRenderer')
    def test_translate_error_reaches_error_policy(self, mock_renderer):
        error_policy = MagicMock()
        mock_renderer.render.side_effect = Exception
        mytx = self._get_tx(error_policy=error_policy)
        mytx.translate('My String', 'en', is_source=False)
        error_policy.get.assert_called_once_with(
            source_string='My String', translation=None, language_code='en',
            escape=True, params={},
        )

    def test_translate_error_reaches_source_string_error_policy(
        self
    ):
        # Trigger a random error in rendering to fallback to the
        # error policy, e.g. an error in missing_policy
        mock_missing_policy = MagicMock()
        mock_missing_policy.get.side_effect = Exception
        mytx = self._get_tx(missing_policy=mock_missing_policy)
        result = mytx.translate('My String', 'en', is_source=False)
        assert result == 'My String'

    @patch('transifex.native.core.StringRenderer')
    @patch('transifex.native.rendering.StringRenderer')
    def test_source_string_policy_custom_text(
        self, mock_renderer1, mock_renderer2
    ):
        error_policy_identifier = (
            'transifex.native.rendering.SourceStringErrorPolicy',
            {'default_text': 'my-default-text'}
        )
        error_policy = parse_error_policy(error_policy_identifier)

        mock_renderer1.render.side_effect = Exception
        mock_renderer2.render.side_effect = Exception
        mytx = self._get_tx(
            error_policy=error_policy
        )
        result = mytx.translate('My String', 'en', is_source=False)
        assert result == 'my-default-text'

    def test_translate_source_language_renders_icu(self):
        mytx = self._get_tx()
        translation = mytx.translate(
            u'{cnt, plural, one {{cnt} duck} other {{cnt} ducks}}',
            'en',
            is_source=True,
            params={'cnt': 1}
        )
        assert translation == '1 duck'

    @patch('transifex.native.core.MemoryCache.get')
    def test_translate_target_language_renders_icu(self, mock_cache):
        mock_cache.return_value = u'{cnt, plural, one {{cnt} παπί} other {{cnt} παπιά}}'
        mytx = self._get_tx()
        translation = mytx.translate(
            u'{cnt, plural, one {{cnt} duck} other {{cnt} ducks}}',
            'en',
            is_source=False,
            params={'cnt': 1}
        )
        assert translation == u'1 παπί'

    def test_translate_source_language_escape_html_true(self):
        mytx = self._get_tx()
        translation = mytx.translate(
            u'<script type="text/javascript">alert(1)</script>',
            'en',
            is_source=True,
            escape=True,
            params={'cnt': 1}
        )
        assert translation == (
            u'&lt;script type=&quot;text/javascript&quot;&gt;alert(1)&lt;/script&gt;'
        )

    def test_translate_source_language_escape_html_false(self):
        mytx = self._get_tx()
        translation = mytx.translate(
            u'<script type="text/javascript">alert(1)</script>',
            'en',
            is_source=True,
            escape=False,
            params={'cnt': 1}
        )
        assert translation == u'<script type="text/javascript">alert(1)</script>'

    def test_translate_source_language_returns_overridden_translation(self):
        """If there is a newer translation in the cache for the source language
        use that instead of the original source string.
        """
        # Populate the cache with two strings in the source language,
        # essentially providing custom translations for the source strings
        cache = MemoryCache()
        source_string1 = u'{cnt, plural, one {{cnt} duck} other {{cnt} ducks}}'
        updated_string1 = u'{???, plural, one {# goose} other {# geese}}'
        key1 = generate_key(source_string1)
        source_string2 = u'Hello {username}'
        updated_string2 = u'Hello to you, {username}!'
        key2 = generate_key(source_string2)
        data = {
            'en': (
                True, {
                    key1: {'string': updated_string1},
                    key2: {'string': updated_string2},
                }
            )
        }
        cache.update(data)

        # Make sure the updated translations are returned
        mytx = self._get_tx(cache=cache)
        translation = mytx.translate(
            source_string1,
            'en',
            is_source=True,
            params={'cnt': 1}
        )
        assert translation == u'1 goose'
        translation = mytx.translate(
            source_string1,
            'en',
            is_source=True,
            params={'cnt': 10}
        )
        assert translation == u'10 geese'
        translation = mytx.translate(
            source_string2,
            'en',
            is_source=True,
            params={'username': 'Jane'}
        )
        assert translation == u'Hello to you, Jane!'

    def test_translate_source_language_returns_original_string(self):
        """If there is no newer translation in the cache for the source language
        use the original source string.
        """
        source_string1 = u'{cnt, plural, one {{cnt} duck} other {{cnt} ducks}}'
        source_string2 = u'Hello {username}'

        # Make sure the updated translations are returned
        mytx = self._get_tx()
        translation = mytx.translate(
            source_string1,
            'en',
            is_source=True,
            params={'cnt': 1}
        )
        assert translation == u'1 duck'
        translation = mytx.translate(
            source_string1,
            'en',
            is_source=True,
            params={'cnt': 10}
        )
        assert translation == u'10 ducks'
        translation = mytx.translate(
            source_string2,
            'en',
            is_source=True,
            params={'username': 'Jane'}
        )
        assert translation == u'Hello Jane'

    def test_translate_with_custom_keys(self):
        # Populate the cache with two strings in the source language,
        # essentially providing custom translations for the source strings
        cache = MemoryCache()
        source_string1 = u'Table'
        source_string2 = u'Hello {username}'
        data = {
            'en': (
                True, {
                    'mykey_1': {'string': u'A table'},
                    'mykey_2': {'string': u'Hello {username}!'},
                }
            ),
            'el': (
                True, {
                    'mykey_1': {'string': u'Ένα τραπέζι'},
                    'mykey_2': {'string': u'Γεια σου, {username}!'},
                }
            )
        }
        cache.update(data)
        mytx = self._get_tx(cache=cache)

        translation = mytx.translate(
            source_string1,
            'en',
            is_source=True,
            _key='mykey_1',
        )
        assert translation == u'A table'
        translation = mytx.translate(
            source_string1,
            'el',
            _key='mykey_1',
        )
        assert translation == u'Ένα τραπέζι'

        translation = mytx.translate(
            source_string2,
            'en',
            is_source=True,
            params={'username': 'Jane'},
            _key='mykey_2',
        )
        assert translation == u'Hello Jane!'
        translation = mytx.translate(
            source_string2,
            'el',
            params={'username': 'Jane'},
            _key='mykey_2',
        )
        assert translation == u'Γεια σου, Jane!'

    @patch('transifex.native.core.CDSHandler.get_push_status')
    @patch('transifex.native.core.CDSHandler.push_source_strings')
    def test_push_strings_reaches_cds_handler(
        self, mock_push_strings, mock_get_status
    ):
        response = MagicMock()
        response.status_code = 202
        response.content = '{"data":{"links":{"job":"/job"}}}'
        mock_push_strings.return_value = response

        response = MagicMock()
        response.status_code = 200
        response.content = '{"data":{"status":"completed","details":{}}}'
        mock_get_status.return_value = response

        strings = [SourceString('a'), SourceString('b')]
        mytx = self._get_tx()
        mytx.push_source_strings(strings, False)
        mock_push_strings.assert_called_once_with(strings, False)

    @patch('transifex.native.core.MemoryCache.update')
    @patch('transifex.native.core.CDSHandler.fetch_translations')
    def test_fetch_translations_reaches_cds_handler_and_cache(self, mock_cds,
                                                              mock_cache):
        mytx = self._get_tx()
        mytx.fetch_translations()
        assert mock_cds.call_count == 1
        assert mock_cache.call_count > 0

    @patch('transifex.native.core.MemoryCache.get')
    def test_plural(self, cache_mock):
        cache_mock.return_value = u'{???, plural, one {ONE} other {OTHER}}'
        tx = self._get_tx()
        translation = tx.translate(u'{cnt, plural, one {one} other {other}}',
                                   u"fr_FR",
                                   params={'cnt': 1})
        assert translation == u'ONE'
        translation = tx.translate(u'{cnt, plural, one {one} other {other}}',
                                   u"fr_FR",
                                   params={'cnt': 2})
        assert translation == u'OTHER'



from __future__ import absolute_import, unicode_literals

from copy import copy

from django.conf import settings
from django.template import Library, Node, TemplateSyntaxError
from django.template.base import (BLOCK_TAG_END, BLOCK_TAG_START,
                                  COMMENT_TAG_END, COMMENT_TAG_START,
                                  VARIABLE_TAG_END, VARIABLE_TAG_START)
from django.template.defaulttags import token_kwargs
from django.utils.html import escape as escape_html
from django.utils.safestring import SafeData, mark_safe
from django.utils.translation import get_language, to_locale
from transifex.common._compat import string_types
from transifex.native import tx
from transifex.native.django.compat import (TOKEN_BLOCK, TOKEN_COMMENT,
                                            TOKEN_TEXT, TOKEN_VAR)

from .utils import get_icu_keys

try:
    from django.utils.safestring import EscapeData
except ImportError:
    class EscapeData(object):
        pass


register = Library()


def do_t(parser, token):
    """ Parse a transifex translation tag.

        Inline syntax:
          {% t/ut <source>[|filter...] [key=param[|filter...]...] [as <var>] %}

          <source> can either be a literal string or a variable containing a
          string. In each case, the string will be considered an ICU template
          whose parameters will be rendered against the parameters of the tag
          and the context of the template.

        Block syntax:
          {% t/ut [|filter...] [key=param[|filter...]...] [as <var>] %}
            <source>
          {% endt/ut %}

          <source> can be anything. Django template tokens will not be parsed,
          but captured verbatim (which will most likely mess up with ICU
          formatting).

        Arguments:

        :param parser: A django Parser object. Has information about how the
                       templating engine is setup, what filters and tags are
                       available and the tokens that haven't yet been processed

        :param token:  A django Token instance. Has information about the
                       contents of the token we are currently processing
    """

    # We will gradually "consume" bits left-to-right. At any given moment, the
    # leftmost bit is the one we need to be processing
    bits = list(token.split_contents())

    # {% t ... %}
    #    ^
    tag_name = bits.pop(0)

    # not {% t %}
    has_args = bool(bits)
    # {% t |escapejs ... %} ... {% endt %}
    first_arg_is_filter = has_args and bits[0].startswith('|')
    # {% t "hello world" ... %}
    first_arg_is_string_literal = (has_args and
                                   bits[0][0] in ('"', "'") and
                                   bits[0][0] == bits[0][-1])
    # {% t var=var ... %}
    first_arg_is_param = (has_args and
                          not first_arg_is_string_literal and
                          '=' in bits[0])
    # {% t as text %}
    first_arg_is_as = has_args and bits[0] == "as"

    is_block = (not has_args or
                first_arg_is_filter or
                first_arg_is_param or
                first_arg_is_as)

    if not is_block:
        # Treat the source string as a full filter expression. Proper
        # exceptions will be raised if it is not formatted properly
        source_string = parser.compile_filter(bits.pop(0))
    else:
        block = []
        while True:
            try:
                # `next_token` removes the first token from the parser's list
                token = parser.next_token()
            except IndexError:
                # We ran out of tokens
                raise TemplateSyntaxError("'{tag_name}' tag not closed with "
                                          "'end{tag_name}'".
                                          format(tag_name=tag_name))
            if (token.token_type == TOKEN_BLOCK and
                    token.contents == "end{}".format(tag_name)):
                break

            # Here we capture the contents of the block verbatim.
            # Unfortunately, there is some loss of information here: When
            # django captures a token, it saves the content as
            # `text[2:-2].strip()`, which means that the size of the
            # surrounding spaces is lost. This means that we end up capturing
            # `{%     tag%}` as `{% tag %}`
            if token.token_type == TOKEN_VAR:
                block.append(''.join((VARIABLE_TAG_START, ' ', token.contents,
                                      ' ', VARIABLE_TAG_END)))
            elif token.token_type == TOKEN_BLOCK:
                block.append(''.join((BLOCK_TAG_START, ' ', token.contents,
                                      ' ', BLOCK_TAG_END)))
            elif token.token_type == TOKEN_COMMENT and token.contents:
                block.append(''.join((COMMENT_TAG_START, ' ', token.contents,
                                      ' ', COMMENT_TAG_END)))
            elif token.token_type == TOKEN_TEXT:
                block.append(token.contents)
            else:  # pragma: no cover
                # Unreachable code
                raise TemplateSyntaxError("Could not parse body of '{}' tag "
                                          "block".format(tag_name))
        block = ''.join(block)

        # Create a dummy filter expression with the provided filters and then
        # replace its 'var' attribute with the captured text. This effectively
        # makes
        # `{% t %}text{% endt %}` equivalent to `{% t "text" %}` and
        # `{% t |filter %}text{% endt %}` equivalent to `{% t "text"|filter %}`
        # while allowing the text to contain characters that would be invalid
        # with the inline syntax, like newlines and `"`
        if first_arg_is_filter:
            source_string = parser.compile_filter('""' + bits.pop(0))
        else:
            source_string = parser.compile_filter('""')
        source_string.var = block

    # {% t "string" var=value var2=value2 ... %}
    #               ^^^^^^^^^
    params_list = []
    while bits and '=' in bits[0]:
        params_list.append(bits.pop(0))
    # Proper exceptions will be raised if the params are not formatted properly
    params = token_kwargs(params_list, parser, support_legacy=False)

    # {% t "string" as var %}
    #               ^^^^^^
    asvar = None
    if bits:
        if len(bits) != 2 or bits[0] != "as":
            raise TemplateSyntaxError("Unrecognized tail: {}".
                                      format(' '.join(bits)))
        else:
            asvar = bits[1]
            del bits[:2]

    if bits:
        raise TemplateSyntaxError("Unrecognized tail: {}".
                                  format(' '.join(bits)))

    return TNode(tag_name, source_string, params, asvar)


register.tag('t')(do_t)
register.tag('ut')(do_t)


class TNode(Node):
    def __init__(self, tag_name, source_string, params, asvar):
        self.tag_name = tag_name
        self.source_string = source_string
        self.params = params
        self.asvar = asvar

    def render(self, context):
        if isinstance(self.source_string.var, string_types):
            # Tag had a string literal or used block syntax
            source_icu_template = self.source_string.var
        else:
            # We resolve the *variable* of the filter expression, not the
            # expression itself, because we want to apply the filters to the
            # *translation* afterwards. Also, we use autoescape=False in order
            # to ignore django's attempts to escape the source string at this
            # point because we want to use the raw string to look for a
            # translation
            safe_context = copy(context)
            safe_context.autoescape = False
            source_icu_template = self.source_string.var.resolve(safe_context)

        # The values of self.params are filter expressions that can be resolved
        params = {key: value.resolve(context)
                  for key, value in self.params.items()}

        # Perform the translation in two steps: First, we get the translation
        # ICU template. Then we perform ICU rendering against 'params'.
        # In between the two steps, if the tag used was 't' and not 'ut', we
        # perform escaping on the ICU template.
        is_source = get_language() == settings.LANGUAGE_CODE
        locale = to_locale(get_language())  # e.g. from en-us to en_US
        translation_icu_template = tx.get_translation(
            source_string=source_icu_template,
            language_code=locale,
            _context=params.get('_context', None),
            is_source=is_source,
            _key=params.get('_key', None),
        )

        # The ICU template can compile against explicitly passed params as well
        # as any context variable. In order to avoid passing and potentially
        # escaping *every* variable in the context for optimization reasons, we
        # need to filter down to the ones that the ICU template will actually
        # need
        keys = get_icu_keys(source_icu_template)
        if translation_icu_template is not None:
            keys.update(get_icu_keys(translation_icu_template))
        for key in keys:
            if key in params:
                continue
            try:
                params[key] = context[key]
            except KeyError:
                pass

        for key, value in params.items():
            # Django doesn't escape strings until the last moment. For now,
            # escaped strings are "marked" as escaped, using the EscapeData
            # class. Because the low-level transifex toolkit doesn't know about
            # django's escape marking, we perform the escaping manually, if
            # needed.
            should_escape = (
                isinstance(value, string_types) and
                ((context.autoescape and not isinstance(value, SafeData)) or
                 (not context.autoescape and isinstance(value, EscapeData)))
            )
            if should_escape:
                params[key] = escape_html(value)

        if self.tag_name == "t":
            source_icu_template = escape_html(source_icu_template)
            if translation_icu_template is not None:
                translation_icu_template = escape_html(
                    translation_icu_template
                )

        result = tx.render_translation(translation_icu_template, params,
                                       source_icu_template, locale,
                                       escape=False)

        # `self` is not supposed to mutate between invocations of `render`
        # because Django may parse the template once per thread and reuse the
        # nodes between renders ("parse" = "process text into nodes"). To that
        # end, let's safekeep the old value of `self.source_string.var` to put
        # it back in place before returning from render.
        # https://docs.djangoproject.com/en/1.11/howto/custom-template-tags/#thread-safety-considerations  # noqa
        old_source_string_var = self.source_string.var

        # Now we resolve the full source filter expression, after having
        # replaced its text with the outcome of the translation, in order to
        # apply the expression's filters to the translation. The translation is
        # marked as safe in order to prevent further escaping attempts that
        # would introduce the danger of double escaping (eg `<` => `&amp;lt;`)
        self.source_string.var = mark_safe(result)
        result = self.source_string.resolve(context)
        self.source_string.var = old_source_string_var

        if self.asvar is not None:
            # Save the translation outcome to a context variable
            context[self.asvar] = result
            return ""
        else:
            return result


@register.filter
def trimmed(value):
    """ Join a multiline string into a single line

            '\n\n  one\ntwo\n\n  three\n  \n   ' => 'one two three'
    """

    return ' '.join((line.strip()
                     for line in value.splitlines()
                     if line.strip()))



from __future__ import absolute_import, unicode_literals

import json
from copy import deepcopy

import responses
from transifex.api.jsonapi import JsonApi, Resource

from .constants import host
from .payloads import Payloads


class ATestApi(JsonApi):
    HOST = host


@ATestApi.register
class Child(Resource):
    TYPE = "children"


@ATestApi.register
class Parent(Resource):
    TYPE = "parents"


test_api = ATestApi(auth="test_api_key")


child_payloads = Payloads(
    'children', 'child',
    extra={'relationships': {
        'parent': {'data': {'type': "parents", 'id': "1"},
                   'links': {'self': "/children/1/relationships/parent",
                             'related': "/parents/1"}},
    }}
)
parent_payloads = Payloads(
    'parents',
    extra={'relationships': {
        'children': {'links': {'self': "/parents/1/relationships/children",
                               'related': "/parents/1/children"}},
    }}
)


@responses.activate
def test_initialization():
    responses.add(responses.GET, "{}/parents/1".format(host),
                  json={'data': {'type': "parents", 'id': "1"}})
    parents = [test_api.Parent.get('1'),
               test_api.Parent(id='1'),
               {'data': {'type': "parents", 'id': '1'}},
               {'type': "parents", 'id': '1'}]
    children = [test_api.Child(relationships={'parent': parent})
                for parent in parents]
    assert all((children[i] == children[i + 1]
                for i in range(len(children) - 1)))
    assert all((children[i].__dict__ == children[i + 1].__dict__
                for i in range(len(children) - 1)))
    assert all((children[i].parent.__dict__ == children[i + 1].parent.__dict__
                for i in range(len(children) - 1)))

    child = test_api.Child(relationships={'parent': None})
    assert child.relationships == child.related == {'parent': None}


@responses.activate
def test_singular_fetch():
    responses.add(responses.GET, "{}/parents/1".format(host),
                  json={'data': parent_payloads[1]})

    child = test_api.Child(child_payloads[1])

    assert (child.relationships ==
            {'parent': {'data': {'type': "parents", 'id': "1"},
                        'links': {'self': "/children/1/relationships/parent",
                                  'related': "/parents/1"}}})
    assert (child.related['parent'] ==
            child.parent ==
            test_api.Parent(id="1"))
    assert child.parent.attributes == child.parent.attributes == {}

    child.fetch('parent')

    assert len(responses.calls) == 1
    assert (child.related['parent'] ==
            child.parent ==
            test_api.Parent(id="1"))
    assert child.parent.attributes == {'name': "parent 1"}
    assert child.parent.name == "parent 1"


@responses.activate
def test_fetch_plural():
    responses.add(responses.GET, "{}/parents/1/children".format(host),
                  json={'data': child_payloads[1:4],
                        'links': {'next': "/parents/1/children?page=2"}},
                  match_querystring=True)
    responses.add(responses.GET, "{}/parents/1/children?page=2".format(host),
                  json={'data': child_payloads[4:7],
                        'links': {'previous': "/parents/1/children?page=1"}},
                  match_querystring=True)

    parent = test_api.Parent(parent_payloads[1])
    assert 'children' not in parent.related
    parent.fetch('children')
    list(parent.children)

    assert len(responses.calls) == 1
    assert 'children' in parent.related
    assert len(parent.children) == 3
    assert isinstance(parent.children[0], Child)
    assert parent.children[1].id == "2"
    assert parent.children[2].name == "child 3"

    assert parent.children.has_next()
    assert not parent.children.has_previous()
    assert len(list(parent.children.all())) == 6


@responses.activate
def test_change_parent_with_save():
    response_body = deepcopy(child_payloads[1])
    relationship = response_body['relationships']['parent']
    relationship['data']['id'] = 2
    relationship['links']['related'] = relationship['links']['related'].\
        replace('1', '2')

    responses.add(responses.PATCH, "{}/children/1".format(host),
                  json={'data': response_body})

    child = test_api.Child(child_payloads[1])
    child.parent = test_api.Parent(parent_payloads[2])

    assert child.relationships['parent']['data']['id'] == "2"
    assert child.related['parent'].id == child.parent.id == "2"

    child.save()

    assert len(responses.calls) == 1
    call = responses.calls[0]
    assert (json.loads(call.request.body.decode())['data']
            ['relationships']['parent']['data']['id'] ==
            "2")


@responses.activate
def test_change_parent_with_change():
    responses.add(responses.PATCH,
                  "{}/children/1/relationships/parent".format(host))

    child = test_api.Child(child_payloads[1])
    new_parent = test_api.Parent(id="2")
    child.change('parent', new_parent)

    assert child.relationships['parent']['data']['id'] == "2"
    assert child.parent.id == "2"
    assert child.parent == new_parent

    assert len(responses.calls) == 1
    call = responses.calls[0]
    assert (json.loads(call.request.body.decode()) ==
            new_parent.as_relationship())


@responses.activate
def test_add():
    responses.add(responses.POST,
                  "{}/parents/1/relationships/children".format(host))

    parent = test_api.Parent(parent_payloads[1])
    children = [test_api.Child(payload) for payload in child_payloads[1:4]]
    parent.add('children', [children[0],
                            children[1].as_relationship(),
                            children[2].as_resource_identifier()])

    assert len(responses.calls) == 1
    call = responses.calls[0]
    assert (json.loads(call.request.body.decode())['data'] ==
            [{'type': "children", 'id': str(i)} for i in range(1, 4)])



from __future__ import absolute_import, unicode_literals

from copy import deepcopy

import requests

from .collections import Collection
from .utils import (has_data, has_links, is_collection, is_dict, is_fetched,
                    is_list, is_null, is_related, is_related_list, is_resource,
                    is_resource_identifier)


class Resource(object):
    """ Subclass like this:

            >>> class Foo(jsonapi.Resource):
            ...     TYPE = "foos"
            ...     EDITABLE = ['name', 'age', 'parent']

        EDITABLE values can either be names of attributes or relationships.
    """

    TYPE = None
    EDITABLE = None

    # Creation
    def __init__(self, data=None, **kwargs):
        """ Initialize an API resource instance when you know the type. """

        if 'type' in kwargs and kwargs['type'] != self.TYPE:
            raise ValueError("Invalid type")

        if data is not None:
            # Maybe a HTTP response body was passed:
            # - Parent(requests.get('http://api.com/parents/1').json())
            # - Parent(requests.get('http://api.com/parents/1').json()['data'])
            # Or a relationship:
            # - `Parent(user.relationships['parent'])`
            # - `Parent({'data': {'type': "parents", 'id': "1"}})`
            if 'data' in data:
                included = data.get('included')
                data = data['data']
                if included is not None and 'included' not in data:
                    data['included'] = included
            self._overwrite(**data)
        else:
            self._overwrite(**kwargs)

    def _overwrite(self,
                   # Copied from response to the instance
                   id=None, attributes=None, relationships=None, links=None,
                   # Used to overwrite 'related'
                   included=None,
                   # Used in case of redirect responses
                   redirect=None,
                   # Ignored
                   type=None,
                   # Magic
                   **kwargs):
        """ Write to the basic attributes of Resource. Used by '__init__',
            'reload', '__copy__' and 'save'
        """

        # Handle "magic" kwargs
        if attributes is None:
            attributes = {}
        if relationships is None:
            relationships = {}
        for key, value in kwargs.items():
            if is_related(value) or is_related_list(value):
                relationships[key] = value
            else:
                attributes[key] = value

        # Copy from response
        self.id = id

        self.attributes = deepcopy(attributes)

        if links is not None:
            self.links = deepcopy(links)
        else:
            self.links = {}

        self.redirect = redirect

        # Relationships
        self.relationships, self.related = {}, {}
        for key, value in relationships.items():
            self._set_relationship(key, value)
            relationship = self.relationships[key]
            if is_null(relationship) or has_data(relationship):
                self.set_related(key, value)

        if included is not None:
            included = {(item['type'], item['id']): item for item in included}
            for relationship_name, relationship in self.relationships.items():
                if is_null(relationship) or not has_data(relationship):
                    continue
                if is_list(relationship['data']):  # Plural with data
                    new_items = [
                        included.get((r['type'], r['id']),
                                     self.related[relationship_name][i])
                        for i, r in enumerate(relationship['data'])
                    ]
                    self.set_related(relationship_name, new_items)
                else:  # Singular
                    key = (relationship['data']['type'],
                           relationship['data']['id'])
                    if key in included:
                        self.set_related(relationship_name, included[key])

    def _set_relationship(self, key, value):
        """ Set 'value' as 'key' relationship. For value we accept:

            - A Resource object
            - A relationship (a dict with either 'data', 'links' or both)
            - A resource identifier (a dict with 'type' and 'id')
            - A list of a combination of the above
            - A dict with a 'data' field which is a list of a combination of
              the above
            - None

            Regardless, in the end `self.relationships` will resemble an API
            response's relationships.
        """

        if is_related_list(value):
            if has_data(value):
                data = value['data']
            else:
                data = value
            self.relationships[key] = {'data': [
                self.API.as_resource(item).as_resource_identifier()
                for item in data
            ]}
            if has_links(value):
                self.relationships[key]['links'] = value['links']
        elif is_resource(value):
            self.relationships[key] = value.as_relationship()
        else:
            value = deepcopy(value)
            if not is_null(value) and is_resource_identifier(value):
                value = {'data': value}
            if is_null(value) or has_data(value) or has_links(value):
                self.relationships[key] = value
            else:
                raise ValueError("Invalid type '{}' for relationship '{}'".
                                 format(value, key))

    def set_related(self, key, value):
        """ Set 'value' as 'key' relationship's value. Works only with singular
            relationships. For value we accept:

            - A Resource object
            - A JSON representation of a Resource object
            - A full API response of a Resource object
            - A relationship (a dict with a 'data' field)
            - A resource identifier (a dict with 'type' and 'id')
            - A list of a combination of the above
            - A dict with a 'data' field with a list of a combination of the
              above as value
            - None

            Regardless, in the end `self.related[key]` will be a Resource
            instance or None.
        """

        if key not in self.relationships:
            raise ValueError("Cannot change relationship '{}' because it's "
                             "not an existing relationship.".format(key))

        relationship = self.relationships[key]

        if is_list(value) or (has_data(value) and is_list(value['data'])):
            # Plural
            if has_data(value):
                value = value['data']
            self.related[key] = Collection.from_data(self.API, {'data': []})
            for item in value:
                self.related[key].append(self.API.as_resource(item))
            new_relationship = [item.as_resource_identifier()
                                for item in self.related[key]]
            if new_relationship != relationship['data']:
                relationship['data'] = new_relationship
        else:
            # Singular
            value = self.API.as_resource(value)
            null_to_not_null = is_null(relationship) and not is_null(value)
            not_null_to_null = not is_null(relationship) and is_null(value)
            data_changed = (not is_null(relationship) and
                            not is_null(value) and
                            (relationship['data'] !=
                             value.as_resource_identifier()))
            if null_to_not_null or not_null_to_null or data_changed:
                if value is None:
                    self.relationships[key] = None
                else:
                    self.relationships[key] = value.as_relationship()
            self.related[key] = value

    @classmethod
    def as_resource(cls, data):
        """ Little convenience function when we don't know if we are dealing
            with a Resource instance or a dict describing a relationship.
        """

        try:
            return cls(data)
        except Exception:
            return data

    def to_dict(self):
        if self.redirect:
            return self.redirect

        result = {'type': self.TYPE}
        if self.id:
            result['id'] = self.id
        if self.attributes:
            result['attributes'] = self.attributes
        if self.relationships:
            result['relationships'] = self.relationships
        if self.links:
            result['links'] = self.links
        return result

    # Shortcuts
    def __getattr__(self, attr):
        if attr in ('a', 'attributes', 'R', 'relationships', 'r', 'related',
                    'id', 'links', 'redirect', 'API'):
            return super(Resource, self).__getattribute__(attr)
        elif attr in self.attributes:
            return self.attributes[attr]
        elif attr in self.related:
            return self.related[attr]
        else:
            return super(Resource, self).__getattribute__(attr)

    def __setattr__(self, attr, value):
        if attr in ('id', 'attributes', 'relationships', 'related', 'links',
                    'redirect', 'API'):
            super(Resource, self).__setattr__(attr, value)
        elif attr in self.attributes:
            self.attributes[attr] = value
        elif attr in self.relationships:
            try:
                self.set_related(attr, value)
            except ValueError as e:
                raise AttributeError(str(e))
        else:
            super(Resource, self).__setattr__(attr, value)

    # Fetching
    def reload(self, include=None):
        """ Fetch fresh data from the server for the object. """

        params = None
        if include is not None:
            params = {'include': ','.join(include)}
        url = self.links.get('self', self.get_item_url())
        response_body = self.API.request('get', url, params=params)
        if (isinstance(response_body, requests.Response) and
                response_body.status_code == 303):
            self._overwrite(redirect=response_body.headers['Location'])
        else:
            self._overwrite(included=response_body.get('included'),
                            **response_body['data'])

    @classmethod
    def get(cls, id=None, include=None, **filters):
        """ Get a resource object by its ID. """

        if id is not None:
            instance = cls(id=id)
            instance.reload(include=include)
            return instance
        else:
            result = cls.list()
            if include is not None:
                result = result.include(*include)
            return result.get(**filters)

    def fetch(self, *relationship_names, **kwargs):
        """ Fetches 'relationship', if it wasn't included when fetching 'self';
            `force=True` supported. Usage:

                >>> foo.fetch('parent')

            Related object will be available after that:

                >>> print(foo.related['parent'].attributes['name'])

            Supports plural relationships, but only one page will be available:

                >>> foo.fetch('children')
                >>> # Only first page
                >>> print([child.attributes['name']
                ...        for child in foo.related['children']])
                >>> # All pages
                >>> print([child.attributes['name']
                ...        for child in foo.related['children'].all()])

            If only one positional argument is supplied, it will return the
            related object or collection:

                >>> print(child.fetch('parent').name)

                Equivalent to:

                >>> child.fetch('parent')
                >>> print(child.parent.name)
        """

        force = kwargs.pop('force', False)

        for relationship_name in relationship_names:
            if relationship_name not in self.relationships:
                raise ValueError("{} doesn't have relationship '{}'".
                                 format(repr(self), relationship_name))

        for relationship_name in relationship_names:
            relationship = self.relationships[relationship_name]

            if is_null(relationship):
                continue

            related = self.related.get(relationship_name)
            is_singular_fetched = is_fetched(related)
            is_plural_fetched = (is_collection(related) and
                                 all((is_fetched(item) for item in related)))
            if (is_singular_fetched or is_plural_fetched) and not force:
                # Has been fetched already
                continue

            if has_data(relationship) and not is_list(relationship['data']):
                # Singular relationship
                self.related[relationship_name].reload()
            else:
                # Plural relationship
                url = relationship.\
                    get('links', {}).\
                    get('related', "/{}/{}/{}".format(self.TYPE, self.id,
                                                      relationship_name))
                self.related[relationship_name] = Collection(self.API, url)

        if len(relationship_names) == 1:
            # This way you can do `project.fetch('languages').filter(...)`
            return self.related[relationship_names[0]]

    @classmethod
    def list(cls):
        return Collection(cls.API, "/{}".format(cls.TYPE))

    def _collection_method(method):
        def _method(cls, *args, **kwargs):
            return getattr(cls.list(), method)(*args, **kwargs)
        return classmethod(_method)

    filter = _collection_method('filter')
    page = _collection_method('page')
    include = _collection_method('include')
    sort = _collection_method('sort')
    fields = _collection_method('fields')
    extra = _collection_method('extra')
    all_pages = _collection_method('all_pages')
    all = _collection_method('all')

    # Editing
    def save(self, *fields, **kwargs):
        """ For new instances (that have `.id == None`), everything will be
            saved and 'id' and other server-generated fields will be set.

            For existing instances, if `fields` or `cls.EDITABLE` is set, then
            only these fields will be saved.

            Usage:
                >>> class Foo(Resource):
                ...     type = 'foos'
                ...     EDITABLE = ['name']

                >>> foo = Foo.get(1)
                >>> foo.attributes['name'] = 'footastic'
                >>> foo = foo.save()
                >>> # or
                >>> foo = foo.save('name', ...)
        """

        fields = set(fields)

        for key, value in kwargs.items():
            setattr(self, key, value)
            fields.add(key)

        if self.id is not None:
            self._save_existing(*fields)
        else:
            self._save_new(*fields)

    def _save_existing(self, *fields):
        payload = self.as_resource_identifier()
        payload.update(self._generate_data_for_saving(*fields))
        response_body = self.API.request('patch',
                                         self.get_item_url(),
                                         json={'data': payload})
        self._post_save(response_body)

    def _save_new(self, *fields):
        payload = {'type': self.TYPE}
        if self.id is not None:
            payload['id'] = self.id
        payload.update(self._generate_data_for_saving(*fields))
        response_body = self.API.request('post',
                                         self.get_collection_url(),
                                         json={'data': payload})
        self._post_save(response_body)

    def _generate_data_for_saving(self, *fields):
        result = {}
        editable_fields = fields or self.EDITABLE
        if editable_fields is not None:
            for field in editable_fields:
                if field in self.attributes:
                    result.setdefault('attributes', {})[field] =\
                        self.attributes[field]
                elif field in self.relationships:
                    result.setdefault('relationships', {})[field] =\
                        self.relationships[field]
                else:
                    raise ValueError("Unknown field '{}'".format(field))
        else:
            if self.attributes:
                result['attributes'] = self.attributes
            if self.relationships:
                result['relationships'] = self.relationships
        return result

    def _post_save(self, response_body):
        if (isinstance(response_body, requests.Response) and
                response_body.status_code in (202, 204)):
            # Success, but the server did not return any new data so our object
            # is considered sufficiently populated with data
            return

        data = response_body['data']

        related = dict(self.related)
        for relationship_name, related_instance in list(related.items()):
            if is_collection(related_instance):
                continue  # Plural relationship

            try:
                current_id = related_instance.id
            except Exception:
                current_id = None
            try:
                new_id = data['relationships'][relationship_name]['data']['id']
            except Exception:
                new_id = None
            if current_id != new_id:
                if new_id is not None:
                    # Relationship changed, reset
                    related[relationship_name] = self.API.new(
                        data['relationships'][relationship_name]
                    )
                else:
                    # Relationship removed
                    del related[relationship_name]

        relationships = data.pop('relationships', {})
        relationships.update(related)

        self._overwrite(relationships=relationships,
                        included=response_body.get('included'),
                        **data)

    @classmethod
    def create(cls, *args, **kwargs):
        instance = cls(*args, **kwargs)
        instance._save_new()
        return instance

    # Handling files
    @classmethod
    def create_with_form(cls, type=None, **kwargs):
        """ Simply fowrard kwargs to requests, for non
            'application/vnd.api+json' requests (eg file uploads)
        """

        if type is None and cls.TYPE is not None:
            type = cls.TYPE
        response_body = cls.API.request('post',
                                        cls.get_collection_url(),
                                        **kwargs)
        return cls.API.new(response_body)

    def follow(self):
        if self.redirect is None:
            raise ValueError("Cannot follow a non-redirect response")
        response_body = self.API.request('get', self.redirect)
        if is_list(response_body['data']):
            return Collection.from_data(self.API, response_body)
        elif is_dict(response_body['data']):
            return self.API.new(response_body)
        else:  # Unreachable code
            raise ValueError("Unknown format while following redirect")

    def delete(self):
        """ Deletes a resource from the API. Usage:

                >>> foo.delete()
        """

        self.API.request('delete', self.get_item_url())
        self.id = None

    # Editing relationshps
    def change(self, field, value):
        """ Change a singular relationship. Usage:

                >>> # Change `child`'s parent from `parent_a` to `parent_b`
                >>> parent_a, parent_b = Parent.list()[:2]
                >>> child = Child.get(XXX)
                >>> assert child.relationships['parent'] == parent_a
                >>> child.change('parent', parent_b)

            Also works with resource identifiers in case we don't have the full
            Resource instance:

                >>> # Make sure `child_a` and `child_b` have the same parent,
                >>> # without fetching the parent
                >>> child_a, child_b = Child.list()[:2]
                >>> child_b.change('parent',
                ...                child_a.relationships['parent']['data'])

            Note: Depending on the API implementation, this can probably be
            also achieved by changing the relationship and saving:

                >>> # Change `child`'s parent from `parent_a` to `parent_b`
                >>> parent_a, parent_b = Parent.list()[:2]
                >>> child = Child.get(XXX)
                >>> assert (child.relationships['parent']['data']['id'] ==
                ...         parent_a.id)
                >>> child.relationships['parent'] = {
                ...     'data': parent_b.as_resource_identifier(),
                ... }
                >>> child.save('parent')
        """

        value = self.API.as_resource(value)
        self._edit_relationship('patch', field, value.as_resource_identifier())
        self.relationships[field]['data'] = value.as_resource_identifier()
        if self.related[field] != value:
            self.related[field] = value

    def add(self, field, values):
        """ Adds items to a plural relationship. Usage:

                >>> # Lets add 3 new children to `parent`
                >>> parent = Parent.get(XXX)
                >>> child_a, child_b, child_c = Child.list(
                ...     filters={'parent[ne]': parent.id},
                ... )[:3]
                >>> parent.add('children', [child_a, child_b, child_c])

            Also works with resource identifiers in case we don't have the
            Resource instances:

                >>> # Make sure parents of `child_a` and `child_b` become
                ...  # children of `grandparent`
                >>> grandparent.add('children',
                ...                 [child_a.relationships['parent']['data'],
                ...                  child_b.relationships['parent']['data']])

            If the plural relationship was previously fetched, it must be
            refetched for the changes to appear.

                >>> parent.add('children', ...)
                >>> parent.fetch('children', force=True)
        """

        self._edit_plural_relationship('post', field, values)

    def remove(self, field, values):
        """ Removes items from a plural relationship. Usage:

                >>> parent = Parent.get(XXX)
                >>> child_a, child_b = Child.list(
                ...     filters={'parent': parent.id},
                ... )[:2]
                >>> parent.remove('children', [child_a, child_b])

            Also works with resource identifiers in case we don't have the
            Resource instances:

                >>> # Make sure parents of `child_a` and `child_b` are no
                ... # longer children of `grandparent`
                >>> grandparent.remove(
                ...     'children',
                ...     [child_a.relationships['parent']['data'],
                ...      child_b.relationships['parent']['data']]
                ... )

            If the plural relationship was previously fetched, it must be
            refetched for the changes to appear.

                >>> parent.remove('children', ...)
                >>> parent.fetch('children', force=True)
        """

        self._edit_plural_relationship('delete', field, values)

    def reset(self, field, values):
        """ Completely rewrites a plural relationship. Usage:

                >>> parent = Parent.get(XXX)
                >>> child_a, child_b, child_c = Child.list()[:3]
                >>> assert (child_a.relationships['parent']['data']['id'] ==
                ...         parent.id)
                >>> assert (child_b.relationships['parent']['data']['id'] !=
                ...         parent.id)
                >>> assert (child_c.relationships['parent']['data']['id'] !=
                ...         parent.id)

                >>> parent.reset('children', [child_b, child_c])

            If the plural relationship was previously fetched, it must be
            refetched for the changes to appear.

                >>> parent.reset('children', ...)
                >>> parent.fetch('children', force=True)
        """

        self._edit_plural_relationship('patch', field, values)

    def _edit_relationship(self, method, field, value):
        url = self.relationships[field].\
            get('links', {}).\
            get('self',
                "/{}/{}/relationships/{}".format(self.TYPE, self.id, field))
        self.API.request(method, url, json={'data': value})

    def _edit_plural_relationship(self, method, field, values):
        payload = [self.API.as_resource(item).as_resource_identifier()
                   for item in values]
        self._edit_relationship(method, field, payload)

    # Bulk actions
    @classmethod
    def bulk_delete(cls, items):
        """ Delete API resource instances in bulk. The server needs to support
            this using the 'bulk' profile with the
            'application/vnd.api+json;profile="bulk"' Content-Type header.

            Accepts a list of:

              - Full JSON responses representing a resource object
              - Resource objects
              - Resource identifiers
              - Relationships
              - IDs

            Doesn't return anything, but will raise an exception if something
            went wrong.

            Usage:

                >>> foos = Foo.list(...)
                >>> Foo.bulk_delete(foos)
        """

        payload = []

        for item in items:
            item = cls.as_resource(item)
            if not is_resource(item):
                item = cls(id=item)
            payload.append(item.as_resource_identifier())

        cls.API.request('delete',
                        cls.get_collection_url(),
                        json={'data': payload},
                        bulk=True)
        return len(payload)

    @classmethod
    def bulk_create(cls, items):
        """ Create API resource instances in bulk. The server needs to support
            this using the 'bulk' profile with the
            'application/vnd.api+json;profile="bulk"' Content-Type header.

            Accepts a list of:
                - (Unsaved) API resource instances
                - Dictionaries with (optional) 'attributes' and 'relationships'
                  fields
                - 2-tuples of 'attributes', 'relationships'
                - 'attributes'

            Returns a list of the created instances.

            Usage:

                >>> # Only attributes
                >>> result = Foo.bulk_create([{'username': "username1"},
                ...                           {'username': "username2"},
                ...                           {'username': "username3"}])
                >>> result[0].id
                <<< 1
                >>> result[0].attributes['username']
                <<< 'username1'

                >>> # attributes and relationships
                >>> parent = ...
                >>> result = Child.bulk_create([({'username': "username1"},
                ...                              {'parent': parent}),
                ...                             ...])
        """

        payload = []
        for item in items:
            if is_list(item):
                attributes, relationships = item
                item = cls(attributes=attributes, relationships=relationships)
            else:
                item = cls.as_resource(item)
                if not isinstance(item, Resource):
                    item = cls(attributes=item)

            payload.append({'type': cls.TYPE})
            if item.attributes:
                payload[-1]['attributes'] = item.attributes
            if item.relationships:
                payload[-1]['relationships'] = item.relationships
            if item.id:
                payload[-1]['id'] = item.id

        response_body = cls.API.request('post',
                                        cls.get_collection_url(),
                                        json={'data': payload},
                                        bulk=True)
        return Collection.from_data(cls.API, response_body)

    @classmethod
    def bulk_update(cls, items, fields=None):
        """ Update API resource instances in bulk. The server needs to support
            this using the 'bulk' profile with the
            'application/vnd.api+json;profile="bulk"' Content-Type header.

            Accepts a list of:
                - API resource instances (with IDs)
                - Dictionaries with (optional) 'attributes', 'relationships'
                  and (required) 'id' fields
                - 3-tuples of 'id', 'attributes', 'relationships'
                - 2-tuples of 'id', 'attributes'
                - 'ids' (maybe we just want the server to update a timestamp)

            Returns a list of the updated instances.

            Usage:

                >>> foos = Foo.list(...)
                >>> for foo in foos:
                ...     foo.attributes['approved'] = True
                >>> foos = Foo.bulk_update(foos, ['approved'])
        """

        if fields is None:
            fields = cls.EDITABLE

        payload = []
        for item in items:

            if is_list(item):
                try:
                    id, attributes, relationships = item
                except ValueError:
                    id, attributes = item
                    relationships = None
                item = cls(id=id,
                           attributes=attributes,
                           relationships=relationships)
            else:
                item = cls.as_resource(item)
                if not isinstance(item, Resource):
                    item = cls(id=item)

            if item.id is None:
                raise ValueError("'id' not supplied as part of an update "
                                 "operation")

            attributes, relationships = item.attributes, item.relationships
            if fields:
                if attributes is not None:
                    attributes = {key: value
                                  for key, value in attributes.items()
                                  if key in fields}
                if relationships is not None:
                    relationships = {key: value
                                     for key, value in relationships.items()
                                     if key in fields}

            payload.append(item.as_resource_identifier())
            if attributes:
                payload[-1]['attributes'] = attributes
            if relationships:
                payload[-1]['relationships'] = {
                    key: cls.API.as_resource(value).as_relationship()
                    for key, value in relationships.items()
                }

        response_body = cls.API.request('patch',
                                        cls.get_collection_url(),
                                        json={'data': payload},
                                        bulk=True)
        return Collection.from_data(cls.API, response_body)

    # Utils
    def __eq__(self, other):
        other = self.API.as_resource(other)
        return self.as_resource_identifier() == other.as_resource_identifier()

    def __repr__(self):
        if self.__class__ is Resource:
            class_name = "Unknown Resource"
        else:
            class_name = self.__class__.__name__

        if self.id is not None:
            details = self.id
        else:
            details = "Unsaved"

        if self.redirect is not None:
            details += " (redirect ready)"

        if not self.attributes and not self.relationships:
            details += " (Unfetched)"

        return repr("<{}: {}>".format(class_name, details))

    def __copy__(self):
        # Will eventually call `_overwrite` so `deepcopy` will be used
        relationships = deepcopy(self.relationships)
        relationships.update(self.related)
        return self.__class__(id=self.id, attributes=self.attributes,
                              relationships=relationships, links=self.links,
                              redirect=self.redirect)

    def as_resource_identifier(self):
        return {'type': self.TYPE, 'id': self.id}

    def as_relationship(self):
        return {'data': self.as_resource_identifier()}

    @classmethod
    def get_collection_url(cls):
        return "/{}".format(cls.TYPE)

    def get_item_url(self):
        if 'self' in self.links:
            return self.links['self']
        else:
            return "/{}/{}".format(self.TYPE, self.id)



from __future__ import absolute_import, unicode_literals


class JsonApiException(Exception):
    """ Assuming a JSON-API error response generated with:

            >>> def do_something():
            ...     response = requests.get(...)
            ...     if not response.ok:
            ...         raise JsonApiException(response.status_code,
            ...                                response.json()['errors'])
            >>> try:
            ...     do_something()
            ... except JsonApiException as e:
            ...     exc = e

        You can access the exception data like:

            >>> exc.status_code
            <<< 400
            >>> len(exc.errors)
            <<< 3
            >>> exc.errors[0].title
            <<< 'Invalid JSON'
            >>> exc.errors[1].code
            <<< 'Forbidden'
            >>> exc.errors[2].detail
            <<< 'There is already a project with the same name'

        The first error's fields are accessible from the main exception; this
        mostly makes sense if `len(exc.errors) == 1`:

            >>> exc.title
            <<< 'Invalid JSON'
    """

    def __init__(self, status_code, errors):
        errors = [JsonApiError(**error) for error in errors]
        super(JsonApiException, self).__init__(status_code, errors)

    status_code = property(lambda self: self.args[0])
    errors = property(lambda self: self.args[1])

    def to_dict(self):
        return {'errors': [error.to_dict() for error in self.errors]}

    # Shortcuts that make sense if len(errors) == 1
    def first_error_property(field):
        def _get(self):
            return getattr(self.errors[0], field)
        return property(_get)

    status = first_error_property('status')
    code = first_error_property('code')
    title = first_error_property('title')
    detail = first_error_property('detail')
    source = first_error_property('source')


class JsonApiError(object):
    def __init__(self, status, code, title, detail, source=None):
        self.status = status
        self.code = code
        self.title = title
        self.detail = detail
        self.source = source

    def __repr__(self):  # pragma: no cover
        return repr("<JsonApiError: {} - {}>".format(self.code, self.detail))

    def to_dict(self):
        result = {'status': self.status, 'code': self.code,
                  'title': self.title, 'detail': self.detail}
        if self.source is not None:
            result['source'] = self.source
        return result


class NotSingleItem(Exception):
    pass


class DoesNotExist(NotSingleItem):
    pass


class MultipleObjectsReturned(NotSingleItem):
    @property
    def count(self):
        return self.args[0]



from __future__ import absolute_import, unicode_literals

import responses
from transifex.api.jsonapi import JsonApi, Resource
from transifex.api.jsonapi.exceptions import JsonApiException

from .constants import host


class ATestApi(JsonApi):
    HOST = host


@ATestApi.register
class Foo(Resource):
    TYPE = "foos"


test_api = ATestApi(auth="test_api_key")


@responses.activate
def test_exception_during_create():
    responses.add(
        responses.POST,
        "{}/foos".format(host),
        json={'errors': [{
            'status': "409",
            'code': "conflict",
            'title': "Conflict error",
            'detail': "username 'Foo' is already taken",
            'source': {'pointer': "/data/attributes/username"},
        }]},
        status=409,
    )

    exc = None
    try:
        test_api.Foo.create(attributes={'username': "Foo"})
    except JsonApiException as e:
        exc = e

    assert exc.status == "409"
    assert exc.code == "conflict"
    assert exc.title == "Conflict error"
    assert exc.detail == "username 'Foo' is already taken"
    assert exc.source == {'pointer': "/data/attributes/username"}

    assert exc.errors[0].status == "409"
    assert exc.errors[0].code == "conflict"
    assert exc.errors[0].title == "Conflict error"
    assert exc.errors[0].detail == "username 'Foo' is already taken"
    assert exc.errors[0].source == {'pointer': "/data/attributes/username"}



from __future__ import absolute_import, unicode_literals

from copy import deepcopy

import requests
import six

from .auth import BearerAuthentication
from .compat import JSONDecodeError
from .exceptions import JsonApiException
from .resources import Resource

type_ = type  # alias to avoid naming conflicts


class _JsonApiMetaclass(type_):
    """ Simple metaclass that overwrites the `registry` class variable on
        `JsonApi` subclasses. This is so that if multiple API connection types
        are defined on the same application, one won't "contaminate" the other.

            eg

            >>> class API_A(jsonapi.JsonApi):
            ...     HOST = "..."

            >>> @API_A.register
            ... class ResourceA(jsonapi.Resource):
            ...     TYPE = "resource_a"

            >>> class API_AA(API_A):
            ...     HOST = "..."

            >>> @API_AA.register
            ... class ResourceAA(jsonapi.Resource):
            ...     TYPE = "resource_aa"

            >>> class API_B(jsonapi.JsonApi):
            ...     HOST = "..."

            >>> @API_B.register
            ... class ResourceB(jsonapi.Resource):
            ...     TYPE = "resource_b"

            >>> API_A.registry
            <<< [ResourceA]

            >>> API_AA.registry
            <<< [ResourceA, ResourceAA]

            >>> API_B.registry
            <<< [ResourceB]
    """

    def __new__(cls, *args, **kwargs):
        result = super(_JsonApiMetaclass, cls).__new__(cls, *args, **kwargs)

        # Use a copy, not reference to parent's registry
        result.registry = list(getattr(result, 'registry', []))

        return result


class JsonApi(six.with_metaclass(_JsonApiMetaclass, object)):
    """ Inteface for a new {json:api} API connection. Initialization
        parameters:

        - host: The URL of the API
        - headers: A dict of HTTP headers that will be included in every
                   request to the server
        - auth: The authentication method. Can either be:

          1. A callable, whose return value should be a dictionary which will
             be merged with the headers of all HTTP request sent to the API
          2. A string, in which case the 'Authorization' header will be
             `Bearer <auth>`

            >>> class API(jsonapi.JsonApi):
            ...     HOST = "..."
            >>> api = API(host=..., auth=...)

        The arguments are optional and can be edited later with `.setup()`

            >>> api = API()
            >>> api.setup(host=..., auth=...)

        All Resource classes that use this API should be registered to this API
        class:

            >>> @API.register
            ... class Foo(jsonapi.Resource):
            ...     TYPE = "foos"
    """

    HOST = None
    HEADERS = None

    def __init__(self, **kwargs):
        """ Create a new API connection instance. It will use the class's
            registry to build the instance's registries in order to be able to
            lookup API resource classes from their class names or API types.

            Delegates configuration to `setup` method.
        """

        self.type_registry = {}
        self.class_registry = {}

        for base_class in self.__class__.registry:
            # Dynamically create a subclass adding 'self' (the API connection
            # instance) as a class variable to it
            child_class = type_(base_class.__name__,
                                (base_class, ),
                                {'API': self})
            # Lookup the new class by it's name or its TYPE class attribute
            self.type_registry[base_class.TYPE] = child_class
            self.class_registry[base_class.__name__] = child_class

        self.host = self.HOST
        if self.HEADERS is None:
            self.headers = {}
        else:
            self.headers = deepcopy(self.HEADERS)
        self.setup(**kwargs)

    def setup(self, host=None, auth=None, headers=None):
        if host is not None:
            self.host = host

        if auth is not None:
            if callable(auth):
                self.make_auth_headers = auth
            else:
                self.make_auth_headers = BearerAuthentication(auth)

        if headers is not None:
            self.headers.update(headers)

    @classmethod
    def register(cls, klass):
        """ Register a API resource type with this API connection *type* (since
            this is a classmethod). When a new API connection *instance* is
            created (see `__init__`), it will use this to build its own
            registry in order to identify class names or API types with the
            relevant API resource classes.
        """

        cls.registry.append(klass)
        return klass

    def __getattr__(self, attr):
        """ Access a registered API resource class. A class name or API
            resource type can be used.

                >>> class FooApi(JsonApi):
                ...     HOST = "https://api.foo.com"

                >>> @FooApi.register
                ... class Foo(Resource):
                ...     TYPE = "foos"

                >>> foo_api = FooApi()

                >>> foo_api.Foo.list()
                >>> # or
                >>> foo_api.foos.list()
        """

        try:
            return self.class_registry[attr]
        except KeyError:
            try:
                return self.type_registry[attr]
            except KeyError:
                raise AttributeError(attr)

    #                 Required args
    def request(self, method, url,
                # Not passed to requests, used to determine Content-Type
                bulk=False,
                # Forwarded to requests
                headers=None, data=None, files=None,
                allow_redirects=False,
                **kwargs):
        if url.startswith('/'):
            url = "{}{}".format(self.host, url)

        if bulk:
            content_type = 'application/vnd.api+json;profile="bulk"'
        elif (data, files) == (None, None):
            content_type = "application/vnd.api+json"
        else:
            # If data and/or files are set, requests will determine
            # Content-Type on its own
            content_type = None

        actual_headers = dict(self.headers)

        if headers is not None:
            actual_headers.update(headers)
        actual_headers.update(self.make_auth_headers())
        if content_type is not None:
            actual_headers.setdefault('Content-Type', content_type)

        response = requests.request(method, url, headers=actual_headers,
                                    data=data, files=files,
                                    allow_redirects=allow_redirects,
                                    **kwargs)

        if not response.ok:
            try:
                exc = JsonApiException(response.status_code,
                                       response.json()['errors'])
            except Exception:
                response.raise_for_status()
            else:
                raise exc
        try:
            return response.json()
        except JSONDecodeError:
            # Most likely empty response when deleting
            return response

    def new(self, data=None, type=None, **kwargs):
        """ Return a new resource instance, using the appropriate Resource
            subclass, provided that it has been registered with this API
            instance.

                >>> _api = jsonapi.JsonApi(...)
                >>> @_api.register
                ... class Foo(jsonapi.Resource):
                ...     TYPE = "foos"
                >>> obj = _api.new(type="foos", ...)

                >>> isinstance(obj, Foo)
                <<< True
        """

        if data is not None:
            if 'data' in data:
                data = data['data']
            return self.new(**data)
        else:
            if type in self.type_registry:
                klass = self.type_registry[type]
            else:
                # Lets make a new class on the fly
                klass = type_(type.capitalize(),
                              (Resource, ),
                              {'API': self, 'TYPE': type})
            return klass(**kwargs)

    def as_resource(self, data):
        """ Little convenience function when we don't know if we are dealing
            with a Resource instance or a dict describing a relationship. Will
            use the appropriate Resource subclass.
        """

        try:
            return self.new(data)
        except Exception:
            return data



from __future__ import absolute_import, unicode_literals

import datetime


class BearerAuthentication(object):
    def __init__(self, api_key):
        self.api_key = api_key

    def __call__(self):
        return {'Authorization': "Bearer {}".format(self.api_key)}


class ULFAuthentication(object):
    def __init__(self, public, secret=None):
        self.public = public
        self.secret = secret

    def __call__(self):
        if self.secret is None:
            return {'Authorization': "ULF {}".format(self.public)}
        else:
            return {'Authorization': ("ULF {}:{}".
                                      format(self.public, self.secret))}


class JWTAuthentication(object):
    """
        Usage:

            >>> from jsonapi import setup
            >>> setup(host="https://some.host.com",
            ...       auth=JWTAuthentication(payload={'username': "username"},
            ...                              secret="SHARED_SECRET",
            ...                              duration=300))
    """

    def __init__(self, payload, secret, duration, algorithm="HS256",
                 get_now=None):
        self.payload = dict(payload)
        self.secret = secret
        self.duration = duration
        self.algorithm = algorithm

        # Dependency injection for getting the current timestamp; maybe it will
        # make testing easier
        if get_now is None:
            get_now = datetime.datetime.utcnow
        self.get_now = get_now

    def __call__(self):
        import jwt  # Optional requirement, don't require at top level

        exp = self.get_now() + datetime.timedelta(seconds=self.duration)
        payload = dict(self.payload)
        payload['exp'] = exp
        token = jwt.encode(payload=payload,
                           secret=self.secret,
                           algorithm=self.algorithm)
        return {'Authorization': "JWT {}".format(token)}



import time

import transifex

from .jsonapi import JsonApi
from .jsonapi import Resource as JsonApiResource
from .jsonapi.exceptions import JsonApiException


class DownloadMixin(object):
    """ Mixin that offers a download method for Transifex APIv3. """

    @classmethod
    def download(cls, interval=5, *args, **kwargs):
        """ Create and poll an async download. Return the download URL when
            done.
        """

        download = cls.create(*args, **kwargs)
        while True:
            if hasattr(download, 'errors') and len(download.errors) > 0:
                errors = [{'code': e['code'],
                           'detail': e['detail'],
                           'title': e['detail'],
                           'status': '409'}
                          for e in download.errors]
                raise JsonApiException(409, errors)
            if download.redirect:
                return download.redirect
            time.sleep(interval)
            download.reload()


class TransifexApi(JsonApi):
    HOST = "https://rest.api.transifex.com"
    HEADERS = {
        'User-Agent': "Transifex-API-SDK/{}".format(transifex.__version__),
    }


@TransifexApi.register
class Organization(JsonApiResource):
    TYPE = "organizations"


@TransifexApi.register
class Team(JsonApiResource):
    TYPE = "teams"


@TransifexApi.register
class Project(JsonApiResource):
    TYPE = "projects"


@TransifexApi.register
class Language(JsonApiResource):
    TYPE = "languages"


@TransifexApi.register
class Resource(JsonApiResource):
    TYPE = "resources"

    def purge(self):
        count = 0
        # Instead of filter, if Resource had a plural relationship to
        # ResourceString, we could do `self.fetch('resource_strings')`
        for page in list(self.API.ResourceString.
                         filter(resource=self).
                         all_pages()):
            count += len(page)
            self.API.ResourceString.bulk_delete(page)
        return count


@TransifexApi.register
class ResourceString(JsonApiResource):
    TYPE = "resource_strings"


@TransifexApi.register
class ResourceTranslation(JsonApiResource):
    TYPE = "resource_translations"
    EDITABLE = ["strings", 'reviewed', "proofread"]


@TransifexApi.register
class ResourceStringsAsyncUpload(JsonApiResource):
    TYPE = "resource_strings_async_uploads"

    @classmethod
    def upload(cls, resource, content, interval=5):
        """ Upload source content with multipart/form-data.

            :param resource: A (transifex) Resource instance or ID
            :param content: A string or file-like object
            :param interval: How often (in seconds) to poll for the completion
                             of the upload job
        """

        if isinstance(resource, JsonApiResource):
            resource = resource.id

        upload = cls.create_with_form(data={'resource': resource},
                                      files={'content': content})

        while True:
            if hasattr(upload, 'errors') and len(upload.errors) > 0:
                errors = [{
                    'code': e['code'],
                    'detail': e['detail'],
                    'title': e['detail'],
                    'status': '409'} for e in upload.errors]
                raise JsonApiException(409, errors)

            if upload.redirect:
                return upload.follow()
            if (hasattr(upload, 'attributes')
                    and upload.attributes.get("details")):
                return upload.attributes.get("details")

            time.sleep(interval)
            upload.reload()


@TransifexApi.register
class ResourceTranslationsAsyncUpload(JsonApiResource):
    TYPE = "resource_translations_async_uploads"

    @classmethod
    def upload(cls, resource, content, language, interval=5,
               file_type='default'):
        """ Upload translation content with multipart/form-data.

            :param resource: A (transifex) Resource instance or ID
            :param content: A string or file-like object
            :param language: A (transifex) Language instance or ID
            :param interval: How often (in seconds) to poll for the completion
                             of the upload job
            :param file_type: The content file type
        """

        if isinstance(resource, JsonApiResource):
            resource = resource.id

        upload = cls.create_with_form(data={'resource': resource,
                                            'language': language,
                                            'file_type': file_type},
                                      files={'content': content})

        while True:
            if hasattr(upload, 'errors') and len(upload.errors) > 0:
                errors = [{
                    'code': e['code'],
                    'detail': e['detail'],
                    'title': e['detail'],
                    'status': '409'} for e in upload.errors]
                raise JsonApiException(409, errors)

            if upload.redirect:
                return upload.follow()
            if (hasattr(upload, 'attributes')
                    and upload.attributes.get("details")):
                return upload.attributes.get("details")

            time.sleep(interval)
            upload.reload()


@TransifexApi.register
class User(JsonApiResource):
    TYPE = "users"


@TransifexApi.register
class TeamMembership(JsonApiResource):
    TYPE = "team_memberships"


@TransifexApi.register
class ResourceLanguageStats(JsonApiResource):
    TYPE = "resource_language_stats"


@TransifexApi.register
class ResourceStringsAsyncDownload(JsonApiResource, DownloadMixin):
    TYPE = "resource_strings_async_downloads"


@TransifexApi.register
class ResourceTranslationsAsyncDownload(JsonApiResource, DownloadMixin):
    TYPE = "resource_translations_async_downloads"


@TransifexApi.register
class I18nFormat(JsonApiResource):
    TYPE = "i18n_formats"


# This is our global object
transifex_api = TransifexApi()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from mock import patch
from transifex.common.strings import LazyString
from transifex.native.django.utils import lazy_translate, translate


def test_translate_without_translation():
    assert translate('A string') == 'A string'


def test_lazy_translate_without_translation():
    string = lazy_translate('A string')
    assert isinstance(string, LazyString)
    assert string == 'A string'


@patch('transifex.native.core.TxNative.get_translation')
def test_translate_with_translation(mock_get_translation):
    mock_get_translation.return_value = 'Γεια σου, {user}!'
    string = translate("Doesn't matter", user='John')
    assert string == 'Γεια σου, John!'


@patch('transifex.native.core.TxNative.get_translation')
def test_lazy_translate_with_translation(mock_get_translation):
    mock_get_translation.return_value = 'Γεια σου, {user}!'
    string = lazy_translate("Doesn't matter", user='John')
    assert string == 'Γεια σου, John!'



from .apis import JsonApi  # noqa
from .exceptions import (DoesNotExist, JsonApiException,  # noqa
                         MultipleObjectsReturned, NotSingleItem)
from .resources import Resource  # noqa



from __future__ import absolute_import, unicode_literals

import responses
from transifex.api.jsonapi import JsonApi, Resource
from transifex.api.jsonapi.collections import Collection

from .constants import host
from .payloads import Payloads


class ATestApi(JsonApi):
    HOST = host


@ATestApi.register
class Item(Resource):
    TYPE = "items"


@ATestApi.register
class Tag(Resource):
    TYPE = "tags"


test_api = ATestApi(host=host, auth="test_api_key")


payloads = Payloads('items')


@responses.activate
def test_collection():
    responses.add(responses.GET, "{}/items".format(host),
                  json={'data': payloads[1:4]})

    collection = Collection(test_api, '/items')
    list(collection)

    assert len(collection) == 3
    assert isinstance(collection[0], Item)
    assert collection[1].id == "2"
    assert collection[2].name == "item 3"

    assert not collection.has_next()
    assert not collection.has_previous()

    assert list(collection) == list(collection.all())


def test_from_data():
    collection = Collection.from_data(test_api, {'data': payloads[1:4]})

    assert len(collection) == 3
    assert isinstance(collection[0], Item)
    assert collection[1].id == "2"
    assert collection[2].name == "item 3"

    assert not collection.has_next()
    assert not collection.has_previous()

    assert list(collection) == list(collection.all())


@responses.activate
def test_pagination():
    responses.add(responses.GET, "{}/items?page=2".format(host),
                  json={'data': payloads[4:7],
                        'links': {'previous': "/items?page=1"}})

    first_page = Collection.from_data(test_api,
                                      {'data': payloads[1:4],
                                       'links': {'next': "/items?page=2"}})
    assert first_page.has_next()
    second_page = first_page.next()
    list(second_page)

    assert len(responses.calls) == 1

    assert len(second_page) == 3
    assert isinstance(second_page[0], Item)
    assert second_page[1].id == "5"
    assert second_page[2].name == "item 6"

    assert not second_page.has_next()
    assert second_page.has_previous()

    assert list(first_page.all()) == list(first_page) + list(second_page)
    assert ([list(page) for page in first_page.all_pages()] ==
            [list(first_page), list(second_page)])


@responses.activate
def test_all():
    responses.add(responses.GET, "{}/items".format(host),
                  json={'data': payloads[1:4]})
    collection = test_api.Item.list()

    assert len(collection) == 3
    assert isinstance(collection[0], Item)
    assert collection[1].id == "2"
    assert collection[2].name == "item 3"

    assert not collection.has_next()
    assert not collection.has_previous()

    assert list(collection) == list(collection.all())


@responses.activate
def test_all_with_pagination():
    responses.add(responses.GET, "{}/items".format(host),
                  json={'data': payloads[1:4],
                        'links': {'next': "/items?page=2"}},
                  match_querystring=True)
    responses.add(responses.GET, "{}/items?page=2".format(host),
                  json={'data': payloads[4:7],
                        'links': {'previous': "/items?page=1"}},
                  match_querystring=True)

    first_page = test_api.Item.list()

    assert first_page.has_next()
    second_page = first_page.next()
    list(second_page)

    assert len(responses.calls) == 2

    assert len(second_page) == 3
    assert isinstance(second_page[0], Item)
    assert second_page[1].id == "5"
    assert second_page[2].name == "item 6"

    assert not second_page.has_next()
    assert second_page.has_previous()

    assert list(first_page.all()) == list(first_page) + list(second_page)
    assert ([list(page) for page in first_page.all_pages()] ==
            [list(first_page), list(second_page)])


@responses.activate
def test_filter():
    responses.add(responses.GET, "{}/items".format(host),
                  json={'data': payloads[1:5]}, match_querystring=True)
    responses.add(responses.GET, "{}/items?filter[odd]=1".format(host),
                  json={'data': payloads[1:5:2]}, match_querystring=True)

    all_items = test_api.Item.list()
    odd_items = test_api.Item.filter(odd=1)

    assert len(all_items) == 4
    assert len(odd_items) == 2

    assert list(odd_items) == [all_items[0], all_items[2]]

    assert (list(test_api.Item.filter(odd=1)) ==
            list(test_api.Item.list().filter(odd=1)) ==
            list(test_api.Item.filter(odd=2).filter(odd=1)))


@responses.activate
def test_include():
    responses.add(responses.GET, "{}/items".format(host), json={
        'data': [{'type': "items",
                  'id': "1",
                  'relationships': {'tag': {'data': {'type': "tags",
                                                     'id': "1"}}}},
                 {'type': "items",
                  'id': "2",
                  'relationships': {'tag': {'data': {'type': "tags",
                                                     'id': "2"}}}}],
        'included': [{'type': "tags",
                      'id': "1",
                      'attributes': {'name': "tag1"}},
                     {'type': "tags",
                      'id': "2",
                      'attributes': {'name': "tag2"}}],
    })

    item1, item2 = test_api.Item.list()
    assert item1.tag.name == "tag1"
    assert item2.tag.name == "tag2"




PRODUCTION = False
SECRET_KEY = 'fake-key'

LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True


INSTALLED_APPS = [
    'transifex.native.django',
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]



import sys

PYVER = sys.version_info[0]
PY3 = PYVER == 3
PY2 = PYVER == 2

if PY3:
    string_types = (str,)
    text_type = str
    binary_type = bytes
else:
    string_types = basestring,
    text_type = unicode
    binary_type = str



from transifex.common.utils import import_to_python
from transifex.native.cache import AbstractCache
from transifex.native.rendering import (AbstractErrorPolicy,
                                        AbstractRenderingPolicy, ChainedPolicy)


def parse_setting_class(obj):
    # obj is a tuple like (<path>, <params>)
    # or a string like <path>
    try:
        path, params = obj
    except ValueError:
        path, params = obj, None

    _class = import_to_python(path)
    if params:
        return _class(**params)
    return _class()


def parse_rendering_policy(policy):
    """Parse the given rendering policy and return an AbstractRenderingPolicy
    subclass.

    :param Union[AbstractRenderingPolicy, str, tuple(str, dict), list] policy:
        could be
        - an instance of AbstractRenderingPolicy
        - a tuple of the class's path and parameters
        - the class's path
        - a list of AbstractRenderingPolicy objects or tuples or string paths
    :return: an AbstractRenderingPolicy object
    :rtype: AbstractRenderingPolicy
    """
    if isinstance(policy, AbstractRenderingPolicy) or policy is None:
        return policy

    if isinstance(policy, list):
        return ChainedPolicy(*[parse_rendering_policy(p) for p in policy])

    return parse_setting_class(policy)


def parse_error_policy(policy):
    """Parse the given error policy and return an AbstractErrorPolicy
    subclass.

    :param Union[AbstractRenderingPolicy, str, tuple(str, dict)] policy:
        could be
        - an instance of AbstractErrorPolicy
        - a tuple of the class's path and parameters
        - the class's path
    :return: an AbstractErrorPolicy object
    :rtype: AbstractErrorPolicy
    """
    if isinstance(policy, AbstractErrorPolicy) or policy is None:
        return policy

    return parse_setting_class(policy)


def parse_cache(cache):
    """Parse the given cache and return an AbstractCache subclass.

    :param Union[AbstractCache, str, tuple(str, dict)] cache:
        could be
        - an instance of AbstractCache
        - a tuple of the class's path and parameters
        - the class's path
    :return: an AbstractCache object
    :rtype: AbstractCache
    """
    if isinstance(cache, AbstractCache) or cache is None:
        return cache

    return parse_setting_class(cache)



import logging
import sys
import time

import requests
from transifex.native.consts import (KEY_CHARACTER_LIMIT,
                                     KEY_DEVELOPER_COMMENT, KEY_OCCURRENCES,
                                     KEY_TAGS)

TRANSIFEX_CDS_HOST = 'https://cds.svc.transifex.net'

TRANSIFEX_CDS_URLS = {
    'FETCH_AVAILABLE_LANGUAGES': '/languages',
    'FETCH_TRANSLATIONS_FOR_LANGUAGE': '/content/{language_code}',
    'PUSH_SOURCE_STRINGS': '/content/',
    'INVALIDATE_CACHE': '/invalidate',
    'PURGE_CACHE': '/purge',
}

logger = logging.getLogger('transifex.native.cds')
logger.addHandler(logging.StreamHandler(sys.stderr))


# A mapping of meta keys
# (interface_key: cds_key)
# Only contains meta keys that are different between the two
MAPPING = {
    KEY_DEVELOPER_COMMENT: 'developer_comment',
    KEY_CHARACTER_LIMIT: 'character_limit',
    KEY_TAGS: 'tags',
    KEY_OCCURRENCES: 'occurrences',
}

# Number of times to retry connecting to CDS before bailing out
MAX_RETRIES = 3
RETRY_DELAY_SEC = 2


class EtagStore(object):
    """ Manges etags """

    # Probably we need to a duration policy here

    def __init__(self):
        self._mem = {}

    def set(self, key, value):
        self._mem[key] = value

    def get(self, key):
        return self._mem.get(key, '')


class CDSHandler(object):
    """Handles communication with the Content Delivery Service."""

    def __init__(self, configured_languages, token, secret=None,
                 host=TRANSIFEX_CDS_HOST):
        """Constructor.

        :param list configured_languages: a list of language codes for the
            configured languages in the application
        :param str token: the API token to use for connecting to the CDS
        :param str host: the host of the Content Delivery Service
        """
        self.configured_language_codes = configured_languages
        self.token = token
        self.secret = secret
        self.host = host or TRANSIFEX_CDS_HOST
        self.etags = EtagStore()

    def fetch_languages(self):
        """Fetch the languages defined in the CDS for the specific project.

        Contains the source language and all target languages.

        :return: a list of language information
        :rtype: dict
        """

        cds_url = TRANSIFEX_CDS_URLS['FETCH_AVAILABLE_LANGUAGES']
        languages = []

        try:
            response = self.retry_get_request(
                self.host + cds_url,
                headers=self._get_headers(),
            )

            if not response.ok:
                logger.error(
                    'Error retrieving languages from CDS: `{}`'.format(
                        response.reason
                    )
                )
                response.raise_for_status()

            json_content = response.json()
            languages = json_content['data']

        except (KeyError, ValueError):
            # Compatibility with python2.7 where `JSONDecodeError` doesn't
            # exist
            logger.error(
                'Error retrieving languages from CDS: Malformed response')
        except requests.ConnectionError:
            logger.error(
                'Error retrieving languages from CDS: ConnectionError')
        except Exception as e:
            logger.error('Error retrieving languages from CDS: UnknownError '
                         '(`{}`)'.format(str(e)))

        return languages

    def fetch_translations(self, language_code=None):
        """Fetch all translations for the given organization/project/(resource)
        associated with the current token.

        Returns a tuple of refresh flag and a dictionary of the fetched
        translations per language. Refresh flag is going to be True whenever
        fresh data has been acquired, False otherwise.

        :return: a dictionary of (refresh_flag, translations) tuples
        :rtype: dict
        """

        cds_url = TRANSIFEX_CDS_URLS['FETCH_TRANSLATIONS_FOR_LANGUAGE']

        translations = {}

        if not language_code:
            languages = [lang['code'] for lang in self.fetch_languages()]
        else:
            languages = [language_code]

        for language_code in set(languages) & \
                set(self.configured_language_codes):

            try:
                response = self.retry_get_request(
                    (self.host + cds_url.format(language_code=language_code)),
                    headers=self._get_headers(
                        etag=self.etags.get(language_code)
                    )
                )

                if not response.ok:
                    logger.error(
                        'Error retrieving translations from CDS: `{}`'.format(
                            response.reason
                        )
                    )
                    response.raise_for_status()

                # etags indicate that no translation have been updated
                if response.status_code == 304:
                    translations[language_code] = (False, {})
                else:
                    self.etags.set(
                        language_code, response.headers.get('ETag', ''))
                    json_content = response.json()
                    translations[language_code] = (
                        True, json_content['data']
                    )

            except (KeyError, ValueError):
                # Compatibility with python2.7 where `JSONDecodeError` doesn't
                # exist
                logger.error('Error retrieving translations from CDS: '
                             'Malformed response')  # pragma no cover
                translations[language_code] = (False, {})  # pragma no cover
            except requests.ConnectionError:
                logger.error(
                    'Error retrieving translations from CDS: ConnectionError')
                translations[language_code] = (False, {})
            except Exception as e:
                logger.error(
                    'Error retrieving translations from CDS: UnknownError '
                    '(`{}`)'.format(str(e))
                )  # pragma no cover
                translations[language_code] = (False, {})

        return translations

    def push_source_strings(self, strings, purge=False):
        """Push source strings to CDS.

        :param list(SourceString) strings: a list of `SourceString` objects
            holding source strings
        :param bool purge: True deletes destination source content not included
            in pushed content. False appends the pushed content to destination
            source content.
        :return: the HTTP response object
        :rtype: requests.Response
        """
        if not self.secret:
            raise Exception('You need to use `TRANSIFEX_SECRET` when pushing '
                            'source content')

        cds_url = TRANSIFEX_CDS_URLS['PUSH_SOURCE_STRINGS']

        data = {k: v for k, v in [self._serialize(item) for item in strings]}
        try:
            response = requests.post(
                self.host + cds_url,
                headers=self._get_headers(use_secret=True),
                json={
                    'data': data,
                    'meta': {'purge': purge},
                }
            )
            response.raise_for_status()

        except requests.ConnectionError:
            logger.error(
                'Error pushing source strings to CDS: ConnectionError')
        except Exception as e:
            logger.error('Error pushing source strings to CDS: UnknownError '
                         '(`{}`)'.format(str(e)))

        return response

    def get_push_status(self, job_path):
        """Get source string push job status

        :param str job_path: Job url path
        :return: the HTTP response object
        :rtype: requests.Response
        """
        if not self.secret:
            raise Exception('You need to use `TRANSIFEX_SECRET` when polling '
                            'source content push')

        try:
            response = requests.get(
                self.host + job_path,
                headers=self._get_headers(use_secret=True),
            )
            response.raise_for_status()

        except requests.ConnectionError:
            logger.error(
                'Error polling source strings push to CDS: ConnectionError')
        except Exception as e:
            logger.error(
                'Error polling source strings push to CDS: UnknownError '
                '(`{}`)'.format(str(e)))

        return response

    def invalidate_cache(self, purge=False):
        """Invalidate CDS cache.

        :param bool purge: True deletes CDS cache entirely instead of
            triggering a job to re-cache content.
        :return: the HTTP response object
        :rtype: requests.Response
        """
        if not self.secret:
            raise Exception('You need to use `TRANSIFEX_SECRET` when '
                            'invalidating cache')

        cds_url = TRANSIFEX_CDS_URLS['PURGE_CACHE'] if purge else \
            TRANSIFEX_CDS_URLS['INVALIDATE_CACHE']

        try:
            response = requests.post(
                self.host + cds_url,
                headers=self._get_headers(use_secret=True),
                json={}
            )
            response.raise_for_status()

        except requests.ConnectionError:
            logger.error(
                'Error invalidating CDS: ConnectionError')
        except Exception as e:
            logger.error('Error invalidating CDS: UnknownError '
                         '(`{}`)'.format(str(e)))

        return response

    def _serialize(self, source_string):
        """Serialize the given source string to a format suitable for the CDS.

        :param transifex.native.parsing.SourceString source_string: the object
            to serialize
        :return: a tuple that contains ths string key and its data,
            as (key, data)
        :rtype: tuple
        """
        data = {
            'string': source_string.string,
            'meta': {
                MAPPING.get(k, k): v
                for k, v in source_string.meta.items()
            },
        }
        if source_string.context:
            data['meta']['context'] = source_string.context

        return source_string.key, data

    def _get_headers(self, use_secret=False, etag=None):
        """Return the headers to use when making requests.

        :param bool use_secret: if True, the Bearer authorization header
            will also include the secret, otherwise it will only use the token
        :param str etag: an optional etag to include
        :return: a dictionary with all headers
        :rtype: dict
        """
        headers = {
            'Authorization': 'Bearer {token}{secret}'.format(
                token=self.token,
                secret=(':' + self.secret if use_secret else '')
            ),
            'Accept-Encoding': 'gzip',
            'Accept-Version': 'v2',
            'X-NATIVE-SDK': 'python',
        }
        if etag:
            headers['If-None-Match'] = etag

        return headers

    def retry_get_request(self, *args, **kwargs):
        """ Resilient function for GET requests """
        retries, last_response_status = 0, 202
        while (last_response_status == 202 or
                500 <= last_response_status < 600 and
                retries < MAX_RETRIES):

            if 500 <= last_response_status < 600:
                retries += 1
                time.sleep(retries * RETRY_DELAY_SEC)

            response = requests.get(*args, **kwargs)
            last_response_status = response.status_code

        return response



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from transifex.native.tools.migrations.models import (Confidence,
                                                      FileMigration,
                                                      StringMigration)


class TestConfidence(object):
    """Test the Confidence class."""

    def test_to_string(self):
        assert Confidence.to_string(Confidence.HIGH) == 'HIGH'
        assert Confidence.to_string(Confidence.LOW) == 'LOW'


class TestStringMigration(object):
    """Test the StringMigration class."""

    def test_constructor_creates_modified_string(self):
        migration = StringMigration('original', 'new')
        assert migration.modified is True
        assert migration.original == 'original'
        assert migration.new == 'new'
        assert migration.confidence == Confidence.HIGH

    def test_constructor_creates_modified_false_string(self):
        migration = StringMigration('identical_strings', 'identical_strings')
        assert migration.modified is False
        assert migration.original == 'identical_strings'
        assert migration.new == 'identical_strings'
        assert migration.confidence == Confidence.HIGH

    def test_update_append(self):
        migration = StringMigration('first', 'FIRST', Confidence.HIGH)
        migration.update('\nsecond', '\nSECOND')
        assert migration.confidence == Confidence.HIGH

        migration.update('\nthird', '\nTHIRD', Confidence.LOW)
        assert migration.confidence == Confidence.LOW
        assert migration.modified is True
        assert migration.original == 'first\nsecond\nthird'
        assert migration.new == 'FIRST\nSECOND\nTHIRD'

    def test_update_prepend(self):
        migration = StringMigration('first', 'FIRST', Confidence.HIGH)
        migration.update('\nsecond', '\nSECOND', append=False)
        migration.update('\nthird', '\nTHIRD', Confidence.LOW, append=False)
        assert migration.confidence == Confidence.LOW
        assert migration.modified is True
        assert migration.original == '\nthird\nsecondfirst'
        assert migration.new == '\nTHIRD\nSECONDFIRST'

    def test_revert(self):
        migration = StringMigration('original', 'new')
        migration.revert()
        assert migration.modified is False
        assert migration.original == 'original'
        assert migration.new == 'original'


class TestFileMigration(object):
    """Test the FileMigration class."""

    def test_constructor(self):
        migration = FileMigration('filename', 'doesnt-matter')
        assert migration.filename == 'filename'
        assert migration.original_content == 'doesnt-matter'
        assert migration.compile() == ''
        assert migration.modified_strings == []

    def test_add_strings_compiles_fine(self):
        migration = FileMigration('filename', 'doesnt-matter')

        m1 = StringMigration('first', 'πρώτο')
        m2 = StringMigration('\nsecond', '\nδεύτερο')
        m3 = StringMigration('\nêà', '\nεα')
        migration.add_string(m1)
        migration.add_string(m2)
        migration.add_string(m3)
        assert migration.modified_strings == [m1, m2, m3]
        assert migration.compile() == 'πρώτο\nδεύτερο\nεα'

    def test_revert_strings(self):
        migration = FileMigration('filename', 'doesnt-matter')

        m1 = StringMigration('first', 'πρώτο')
        m2 = StringMigration('second', 'δεύτερο')
        m3 = StringMigration('third', 'τρίτο')
        migration.add_string(m1)
        migration.add_string(m2)
        migration.add_string(m3)
        migration.revert()
        assert migration.compile() == 'firstsecondthird'
        assert migration.modified_strings == []



from setuptools import find_packages, setup

import versioneer

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name='transifex-python',
    author='Transifex',
    author_email='info@transifex.com',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description=(
        'Transifex Python Toolkit'
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    url='https://github.com/transifex/transifex-python',
    install_requires=[
        'pyseeyou', 'pytz', 'requests', 'click', 'asttokens'
    ],
)



# -*- coding: utf-8 -*-
import pytest
from mock import patch
from transifex.native.rendering import (ChainedPolicy, ExtraLengthPolicy,
                                        PseudoTranslationPolicy,
                                        SourceStringErrorPolicy,
                                        SourceStringPolicy, StringRenderer,
                                        WrappedStringPolicy)
from transifex.native.settings import parse_rendering_policy

COMPLEX_STRINGS = u"""{gender_of_host, select,
  female {
    {total_guests, plural, offset:1
      =0 {{host} does not give a party.}
      =1 {{host} invites {guest} to her party.}
      =2 {{host} invites {guest} and one other person to her party.}
      other {{host} invites {guest} and # other people to her party.}}}
  male {
    {total_guests, plural, offset:1
      =0 {{host} does not give a party.}
      =1 {{host} invites {guest} to his party.}
      =2 {{host} invites {guest} and one other person to his party.}
      other {{host} invites {guest} and # other people to his party.}}}
  other {
    {total_guests, plural, offset:1
      =0 {{host} does not give a party.}
      =1 {{host} invites {guest} to their party.}
      =2 {{host} invites {guest} and one other person to their party.}
      other {{host} invites {guest} and # other people to their party.}}}
}"""

JS_SCRIPT = u'<script type="text/javascript">alert(1)</script>'


class TestStringRenderer(object):
    """Tests the functionality of the StringRenderer class."""

    def test_simple_render_escaped(self):
        translation = StringRenderer.render(
            JS_SCRIPT,
            JS_SCRIPT,
            'en',
            escape=True,
            missing_policy=SourceStringPolicy(),
            params={'cnt': 2},
        )
        assert translation == (
            u'&lt;script type=&quot;text/javascript&quot;&gt;alert(1)&lt;/script&gt;'
        )

    def test_simple_render_unescaped(self):
        translation = StringRenderer.render(
            JS_SCRIPT,
            JS_SCRIPT,
            'en',
            escape=False,
            missing_policy=SourceStringPolicy(),
            params={'cnt': 2},
        )
        assert translation == JS_SCRIPT

    def test_simple_render_with_translation(self):
        translation = StringRenderer.render(
            u'{cnt, plural, one {{cnt} table} other {{cnt} tables}}',
            u'{cnt, plural, one {{cnt} τραπέζι} other {{cnt} τραπέζια}}',
            'en',
            escape=True,
            missing_policy=SourceStringPolicy(),
            params={'cnt': 2},
        )
        assert translation == u'2 τραπέζια'

    def test_simple_render_with_missing_translation(self):
        translation = StringRenderer.render(
            u'{cnt, plural, one {{cnt} table} other {{cnt} tables}}',
            None,
            'en',
            escape=True,
            missing_policy=SourceStringPolicy(),
            params={'cnt': 2},
        )
        # Should fall back to source
        assert translation == u'2 tables'

        translation = StringRenderer.render(
            u'{cnt, plural, one {{cnt} table} other {{cnt} tables}}',
            None,
            'en',
            escape=True,
            missing_policy=PseudoTranslationPolicy(),
            params={'cnt': 2},
        )
        # Should use the proper missing policy
        assert translation == u'2 ťàƀĺêš'

    def test_simple_render_escaped_with_missing_translation(self):
        translation = StringRenderer.render(
            JS_SCRIPT,
            None,
            'en',
            escape=True,
            missing_policy=SourceStringPolicy(),
            params={'cnt': 2},
        )
        assert translation == (
            u'&lt;script type=&quot;text/javascript&quot;&gt;alert(1)&lt;/script&gt;'
        )

    def test_simple_render_unescaped_with_missing_translation(self):
        translation = StringRenderer.render(
            JS_SCRIPT,
            None,
            'en',
            escape=False,
            missing_policy=SourceStringPolicy(),
            params={'cnt': 2},
        )
        assert translation == JS_SCRIPT

    def test_complex_message_format(self):
        translation = self._complex(gender_of_host='female', total_guests=1)
        assert translation == u'Jane does not give a party.'

        translation = self._complex(gender_of_host='female', total_guests=2)
        assert translation == u'Jane invites Joe to her party.'

        translation = self._complex(gender_of_host='female', total_guests=3)
        assert translation == u'Jane invites Joe and one other person to her party.'

        translation = self._complex(gender_of_host='female', total_guests=10)
        assert translation == u'Jane invites Joe and 9 other people to her party.'

    def _complex(self, **params):
        params = dict(params)
        params.update({'host': "Jane", 'guest': "Joe"})
        return StringRenderer.render(
            COMPLEX_STRINGS,
            None,
            'en',
            escape=True,
            missing_policy=SourceStringPolicy(),
            params=params,
        )

    @patch('transifex.native.rendering.html_escape')
    @patch('transifex.native.rendering.logger')
    def test_error_raises_exception(self, mock_logger, mock_escape):

        mock_escape.side_effect = Exception
        with pytest.raises(Exception):
            translation = StringRenderer.render(
                'Source String',
                'Translation',
                'en',
                escape=True,
                missing_policy=SourceStringPolicy(),
            )
        mock_logger.error.assert_called_with(
            'RenderingError: Could not render string `%s` in language `%s` '
            'with parameters `%s` (Error: %s, Source String: %s)',
            'Translation', 'en', '{}', '', 'Source String'
        )

    def test_no_missing_policy_and_error_raises_exception(self):
        with pytest.raises(Exception) as exc_info:
            translation = StringRenderer.render(
                'source',
                '',
                'en',
                escape=True,
                missing_policy=None,
            )
        assert str(exc_info.value) == (
            'No string to render and no missing policy defined! '
            '(Source String: `source`)'
        )


class TestMissingPolicies(object):
    """Tests the functionality of the various StringRenderer subclasses
    of AbstractRenderingPolicy."""

    def test_source_string_policy(self):
        assert SourceStringPolicy().get(u'Source-String') == u'Source-String'

    def test_pseudo_translation_policy(self):
        assert PseudoTranslationPolicy().get(u'Source') == u'Șøüȓċê'

    def test_wrapped_string(self):
        missing = WrappedStringPolicy(start='[[', end=']]').get(u'Source')
        assert missing == u'[[Source]]'

        missing = WrappedStringPolicy().get(u'Source')
        assert missing == u'[Source]'

    def test_extra_length(self):
        missing = ExtraLengthPolicy().get(u'Source')
        assert missing == u'Source~e'

        missing = ExtraLengthPolicy(extra_percentage=0.8, extra_str=u'$')\
            .get(u'Source')
        assert missing == u'Source$$$$$'

    def test_chained(self):
        policy = ChainedPolicy(
            PseudoTranslationPolicy(),
            ExtraLengthPolicy(extra_percentage=0.5),
            WrappedStringPolicy(start='>>', end='<<'),
        )
        missing = policy.get(u'This is a long sentence')
        assert missing == u'>>Ťȟıš ıš à ĺøñğ šêñťêñċê~extra~~extr<<'

        # Test again with different order
        policy = ChainedPolicy(
            ExtraLengthPolicy(extra_percentage=0.5),
            WrappedStringPolicy(start='>>', end='<<'),
            PseudoTranslationPolicy(),
        )
        missing = policy.get(u'This is a long sentence')
        assert missing == u'>>Ťȟıš ıš à ĺøñğ šêñťêñċê~êẋťȓà~~êẋťȓ<<'


class TestErrorPolicies(object):
    """Tests the functionality of the various ErrorPolicy subclasses
    of AbstractErrorPolicy."""

    def test_source_string_policy(self):
        assert SourceStringErrorPolicy().get(
            source_string='Source-String',
            translation=None,
            language_code='a',
            escape=True
        ) == 'Source-String'

    @patch('transifex.native.rendering.logger')
    @patch('transifex.native.rendering.StringRenderer')
    def test_source_string_policy(self, mock_renderer, mock_logger):
        mock_renderer.render.side_effect = Exception
        assert SourceStringErrorPolicy().get(
            source_string='Source-String',
            translation=None,
            language_code='a',
            escape=True
        ) == 'ERROR'
        mock_logger.error.assert_called_with(
            'ErrorPolicyError: Could not render string `Source-String` with parameters `{}`'
        )


class TestParsing(object):
    """Tests the parse_rendering_policy() method."""

    def test_none_returns_none(self):
        assert parse_rendering_policy(None) is None

    def test_policy_object_returns_the_object(self):
        policy = WrappedStringPolicy()
        assert parse_rendering_policy(policy) == policy

    def test_list(self):
        """Test a list of policy objects, string paths and path/param tuples."""
        parsed = parse_rendering_policy([
            WrappedStringPolicy('{{', '}}'),
            'transifex.native.rendering.PseudoTranslationPolicy',
            (
                'transifex.native.rendering.ExtraLengthPolicy',
                {'extra_percentage': 0.5, 'extra_str': 'foo'},
            )
        ])
        assert isinstance(parsed, ChainedPolicy)
        assert len(parsed._policies) == 3
        assert parsed._policies[0].__dict__ == {'end': '}}', 'start': '{{'}
        assert isinstance(parsed._policies[1], PseudoTranslationPolicy)
        assert parsed._policies[2].__dict__ == {
            'extra_percentage': 0.5, 'extra_str': 'foo',
        }



from __future__ import absolute_import, unicode_literals

import six

from .compat import abc


def is_resource(value):
    from .resources import Resource
    return isinstance(value, Resource)


def is_collection(value):
    from .collections import Collection
    return isinstance(value, Collection)


def is_dict(value):
    return isinstance(value, abc.Mapping)


def is_list(value):
    return (isinstance(value, abc.Sequence) and
            not isinstance(value, six.string_types))


def is_null(value):
    return value is None


def has_data(value):
    return is_dict(value) and 'data' in value


def has_links(value):
    return is_dict(value) and 'links' in value


def is_resource_identifier(value):
    return is_dict(value) and {'type', 'id'} <= set(value.keys())


def is_relationship(value):
    return (is_dict(value) and
            has_data('value') and
            is_dict(value['data']) and
            is_resource_identifier(value['data']))


def is_related(value):
    """ Determines if value can be considered a relationship in any way. """

    return (is_resource(value) or
            is_resource_identifier(value) or
            is_relationship(value))


def is_related_list(value):
    if has_data(value):
        value = value['data']

    return is_list(value) and all((is_related(item) for item in value))


def is_fetched(value):
    return is_resource(value) and (value.attributes or value.relationships)



from __future__ import absolute_import, unicode_literals

from transifex.api.jsonapi import JsonApi, Resource
from transifex.api.jsonapi.auth import ULFAuthentication

from .constants import host


class ATestApi(JsonApi):
    HOST = host


@ATestApi.register
class GlobalTest(Resource):
    TYPE = "globaltests"


test_api = ATestApi()


def reset_setup():
    test_api.setup(host=host, auth="test_api_key")
    assert (test_api.make_auth_headers() ==
            {'Authorization': "Bearer test_api_key"})
    assert test_api.host == host


def test_registries():
    assert issubclass(test_api.type_registry['globaltests'], GlobalTest)
    assert issubclass(test_api.class_registry['GlobalTest'], GlobalTest)


def test_setup_plaintext():
    test_api.setup(host="http://some.host", auth="another_key")
    assert (test_api.make_auth_headers() ==
            {'Authorization': "Bearer another_key"})
    assert test_api.host == "http://some.host"
    reset_setup()


def test_setup_ulf():
    test_api.setup(auth=ULFAuthentication('public'))
    assert test_api.make_auth_headers() == {'Authorization': "ULF public"}

    test_api.setup(auth=ULFAuthentication('public', 'secret'))
    assert (test_api.make_auth_headers() ==
            {'Authorization': "ULF public:secret"})

    reset_setup()


def test_setup_any_callable():
    test_api.setup(host="http://some.host2",
                   auth=lambda: {'Authorization': "Another key2"})
    assert test_api.make_auth_headers() == {'Authorization': "Another key2"}
    assert test_api.host == "http://some.host2"
    reset_setup()
